#!/usr/bin/env python


import argparse
import collections
import copy
import heapq
import json
import logging
import os
import re
import sqlite3
import uuid

import numpy as np

import neuroglancer
import neuroglancer.cli

debug_graph = False
verbose_merging = False


def normalize_edge(e):
    id_a, id_b = e
    if id_a > id_b:
        id_a, id_b = id_b, id_a
    return id_a, id_b


class GreedyMulticut:
    def __init__(self, combine_edges, edge_priority):
        # Contains (score, edge_map_value) tuple values in heap order.  The
        # edge_map_value is the actual corresponding value in edge_map, not a copy.
        self.edge_heap = []

        # Maps segment_id -> set of segment_id neighbors.
        self.regions = dict()

        # Maps (id_a, id_b) -> edge_map_value=[score, key, edge_object]
        self.edge_map = dict()
        self.combine_edges = combine_edges
        self.edge_priority = edge_priority
        self.num_valid_edges = 0
        self._initialized = False

    def add_edge(self, e, edge):
        id_a, id_b = e
        id_a, id_b = normalize_edge((id_a, id_b))
        self.regions.setdefault(id_a, set()).add(id_b)
        self.regions.setdefault(id_b, set()).add(id_a)
        key = (id_a, id_b)
        entry = self.edge_map.get(key, None)
        if entry is not None:
            edge_data = entry[2] = self.combine_edges(entry[0], edge)
            entry[0] = self.edge_priority(edge_data)
        else:
            entry = self.edge_map[key] = [self.edge_priority(edge), key, edge]
            self.num_valid_edges += 1
        if self._initialized:
            self._add_to_heap(entry)

    def _initialize_heap(self):
        if self._initialized:
            return
        for key in self.edge_map:
            entry = self.edge_map[key]
            self._add_to_heap(entry)
        self._initialized = True

    def _add_to_heap(self, entry):
        heapq.heappush(self.edge_heap, (entry[0], entry))

    def remove_edge_from_heap(self, segment_ids):
        """Remove an edge from the heap."""
        self._initialize_heap()
        key = normalize_edge(segment_ids)
        if key in self.edge_map:
            self.edge_map[key][0] = None
            self.num_valid_edges -= 1

    def check_consistency(self):
        self._initialize_heap()
        expected_regions = dict()
        for key, entry in self.edge_map.items():
            assert entry[1] == key
            expected_regions.setdefault(key[0], set()).add(key[1])
            expected_regions.setdefault(key[1], set()).add(key[0])

        assert expected_regions == self.regions

        num_valid_edges = 0
        for e in self.edge_heap:
            if self._is_valid_heap_entry(e):
                num_valid_edges += 1
        assert num_valid_edges == self.num_valid_edges

    def merge(self, e):
        id_a, id_b = e
        self._initialize_heap()
        id_a, id_b = normalize_edge((id_a, id_b))
        if (id_a, id_b) not in self.edge_map:
            raise KeyError
        for neighbor in self.regions[id_b]:
            if neighbor == id_a:
                continue
            expired_ids = normalize_edge((neighbor, id_b))
            new_ids = normalize_edge((neighbor, id_a))
            new_edge = self.edge_map.get(new_ids)
            expired_edge = self.edge_map[expired_ids]
            if new_edge is not None:
                edge_data = new_edge[2] = self.combine_edges(
                    new_edge[2], expired_edge[2]
                )
                if new_edge[0] is not None:
                    self.num_valid_edges -= 1
                if expired_edge[0] is not None:
                    self.num_valid_edges -= 1
                self.num_valid_edges += 1
                new_edge[0] = self.edge_priority(edge_data)
                self._add_to_heap(new_edge)
            else:
                self.regions[neighbor].add(id_a)
                self.regions[id_a].add(neighbor)
                self.edge_map[new_ids] = expired_edge
                expired_edge[1] = new_ids
                # No need to add to heap, since score hasn't changed.
            del self.edge_map[expired_ids]
            self.regions[neighbor].remove(id_b)
        del self.regions[id_b]
        self.regions[id_a].remove(id_b)
        del self.edge_map[(id_a, id_b)]
        self.num_valid_edges -= 1

    def _is_valid_heap_entry(self, heap_entry):
        score, entry = heap_entry
        expected_entry = self.edge_map.get(entry[1])
        if entry is not expected_entry or entry[0] is not score:
            return None
        return entry

    def get_next_edge(self):
        self._initialize_heap()
        while True:
            if self.num_valid_edges == 0:
                return None
            heap_entry = self.edge_heap[0]
            entry = self._is_valid_heap_entry(heap_entry)
            if entry is None:
                heapq.heappop(self.edge_heap)
            else:
                return entry


Edge = collections.namedtuple("Edge", ["segment_ids", "score", "position"])


def load_edges(path):
    edges = []
    with open(path) as f:
        f.readline()
        for line in f:
            parts = line.split(",")
            segment_a = int(parts[0].strip())
            segment_b = int(parts[1].strip())
            score = float(parts[2].strip())
            position = (
                int(parts[3].strip()),
                int(parts[4].strip()),
                int(parts[5].strip()),
            )
            edges.append(
                Edge(segment_ids=(segment_a, segment_b), score=score, position=position)
            )
    return edges


def load_split_seeds(path):
    with open(path) as f:
        raw_seeds = json.loads(f.read())
    seeds = {}
    for component in raw_seeds:
        seeds.setdefault(component["label"], []).extend(component["supervoxels"])
    return seeds


def build_graph(edges):
    logging.info("Building graph with %d edges", len(edges))

    def combine_edges(a, b):
        return a + b

    def edge_priority(x):
        return x

    greedy_multicut = GreedyMulticut(
        combine_edges=combine_edges,
        edge_priority=edge_priority,
    )
    for edge in edges:
        greedy_multicut.add_edge(edge.segment_ids, edge.score)
    return greedy_multicut


class AgglomerationGraph:
    def __init__(self, conn):
        self.conn = conn
        self.agglo_members_cache = dict()
        self.agglo_edges_cache = dict()

    def get_agglo_id(self, supervoxel_id):
        c = self.conn.cursor()
        c.execute(
            "SELECT agglo_id FROM supervoxels WHERE supervoxel_id=?",
            (int(supervoxel_id),),
        )
        result = c.fetchone()
        if result is None:
            return supervoxel_id
        return result[0]

    def get_agglo_members(self, agglo_id):
        result = self.agglo_members_cache.get(agglo_id)
        if result is not None:
            return result
        c = self.conn.cursor()
        c.execute(
            "SELECT supervoxel_id FROM supervoxels WHERE agglo_id=?", (int(agglo_id),)
        )
        result = [row[0] for row in c.fetchall()]
        self.agglo_members_cache[agglo_id] = result
        return result

    def get_agglo_edges(self, agglo_id):
        result = self.agglo_edges_cache.get(agglo_id)
        if result is not None:
            return result
        c = self.conn.cursor()
        c.execute(
            "SELECT segment_a, segment_b, score, x, y, z FROM edges WHERE agglo_id=?",
            (int(agglo_id),),
        )
        result = [
            Edge(
                segment_ids=(row[0], row[1]),
                score=row[2],
                position=(row[3], row[4], row[5]),
            )
            for row in c.fetchall()
        ]
        self.agglo_edges_cache[agglo_id] = result
        return result


def _make_supervoxel_map(graph, split_seeds, need_agglo_ids):
    supervoxel_map = dict()
    agglo_ids = dict()

    for label in [0, 1]:
        for seed in split_seeds[label]:
            supervoxel_id = seed["supervoxel_id"]
            if need_agglo_ids:
                agglo_id = graph.get_agglo_id(supervoxel_id)
                if agglo_id == 0:
                    continue
                agglo_ids.setdefault(agglo_id, []).append((label, seed))
            supervoxel_map.setdefault(supervoxel_id, set()).add(label)
    return agglo_ids, supervoxel_map


def do_split(graph, split_seeds, agglo_id=None, supervoxels=None):
    agglo_ids, supervoxel_map = _make_supervoxel_map(
        graph, split_seeds, need_agglo_ids=agglo_id is None
    )

    if agglo_id is None:
        agglo_id_counts = {
            agglo_id: sum(z[1]["count"] for z in seeds)
            for agglo_id, seeds in agglo_ids.items()
        }

        agglo_id = max(agglo_ids, key=lambda x: agglo_id_counts[x])

        if len(agglo_ids) > 1:
            logging.info(
                "Warning: more than one agglomerated component.  "
                + "Choosing component %d with maximum number of seed points.",
                agglo_id,
            )
            logging.info("agglo_id_counts = %r", agglo_id_counts)

    input_edges = graph.get_agglo_edges(agglo_id)
    if supervoxels is not None:
        input_edges = [
            x
            for x in input_edges
            if x.segment_ids[0] in supervoxels and x.segment_ids[1] in supervoxels
        ]
    graph = build_graph(input_edges)
    if debug_graph:
        graph.check_consistency()

    cur_eqs = neuroglancer.EquivalenceMap()
    logging.info("Agglomerating")
    threshold = float("inf")
    while True:
        entry = graph.get_next_edge()
        if entry is None:
            if verbose_merging:
                logging.info("Stopping because entry is None")
            break
        if entry[0] > threshold:
            if verbose_merging:
                logging.info(
                    "Stopping because edge score %r is > threshold %r",
                    entry[0],
                    threshold,
                )
            break
        segment_ids = entry[1]
        seeds_a = supervoxel_map.get(segment_ids[0])
        seeds_b = supervoxel_map.get(segment_ids[1])
        if (
            (seeds_a is not None and len(seeds_a) > 1)
            or (seeds_b is not None and len(seeds_b) > 1)
            or (seeds_a is not None and seeds_b is not None and seeds_a != seeds_b)
        ):
            if verbose_merging:
                logging.info(
                    "Excluding edge %r because of seeds: %r %r",
                    segment_ids,
                    seeds_a,
                    seeds_b,
                )
            graph.remove_edge_from_heap(segment_ids)
            continue
        if verbose_merging:
            logging.info("Merging %r with score %r", segment_ids, entry[0])
        graph.merge(segment_ids)
        if debug_graph:
            graph.check_consistency()

        new_id = cur_eqs.union(*segment_ids)
        new_seeds = seeds_a or seeds_b
        if new_seeds:
            supervoxel_map[new_id] = new_seeds

    return dict(agglo_id=agglo_id, cur_eqs=cur_eqs, supervoxel_map=supervoxel_map)


def display_split_result(
    graph, agglo_id, cur_eqs, supervoxel_map, split_seeds, image_url, segmentation_url
):
    agglo_members = set(graph.get_agglo_members(agglo_id))
    state = neuroglancer.ViewerState()
    state.layers.append(name="image", layer=neuroglancer.ImageLayer(source=image_url))
    state.layers.append(
        name="original",
        layer=neuroglancer.SegmentationLayer(
            source=segmentation_url,
            segments=agglo_members,
        ),
        visible=False,
    )
    state.layers.append(
        name="isolated-supervoxels",
        layer=neuroglancer.SegmentationLayer(
            source=segmentation_url,
            segments={x for x, seeds in supervoxel_map.items() if len(seeds) > 1},
        ),
        visible=False,
    )
    state.layers.append(
        name="split",
        layer=neuroglancer.SegmentationLayer(
            source=segmentation_url,
            equivalences=cur_eqs,
            segments={cur_eqs[x] for x in agglo_members},
        ),
    )
    for label, component in split_seeds.items():
        state.layers.append(
            name="seed%d" % label,
            layer=neuroglancer.PointAnnotationLayer(
                points=[seed["position"] for seed in component],
            ),
        )

    state.show_slices = False
    state.layout = "3d"
    all_seed_points = [
        seed["position"] for component in split_seeds.values() for seed in component
    ]
    state.voxel_coordinates = np.mean(all_seed_points, axis=0)
    return state


def _set_viewer_seeds(s, seeds):
    for inclusive in [False, True]:
        layer_name = "inclusive-seeds" if inclusive else "exclusive-seeds"
        s.layers[layer_name] = neuroglancer.AnnotationLayer(
            annotation_color="green" if inclusive else "red",
            annotations=[
                dict(
                    type="point",
                    id=x["id"],
                    point=x["position"],
                    description=str(x["supervoxel_id"]),
                )
                for x in seeds[inclusive]
            ],
        )


def _get_viewer_seeds(s):
    seeds = [[], []]
    for inclusive in [False, True]:
        layer_name = "inclusive-seeds" if inclusive else "exclusive-seeds"
        try:
            layer = s.layers[layer_name]
        except KeyError:
            pass
        for x in layer.annotations:
            seeds[inclusive].append(
                dict(
                    id=x.id,
                    supervoxel_id=int(x.description),
                    position=tuple(map(int, x.point)),
                )
            )
    return seeds


class ComponentState:
    def __init__(self, data=None):
        self.supervoxels = set()
        self.seeds = [[], []]
        if data is not None:
            self.load(data)

    def load(self, data):
        self.supervoxels = set(data["supervoxels"])
        self.seeds = data["seeds"]

    def to_json(self):
        return {
            "supervoxels": sorted(self.supervoxels),
            "seeds": self.seeds,
        }


class InteractiveState:
    def __init__(self, path):
        self.unused_supervoxels = set()
        self.components = []
        self.path = path
        self.selected_component = None

    def load(self):
        with open(self.path) as f:
            data = json.load(f)
            self.unused_supervoxels = set(data["unused_supervoxels"])
            self.components = map(ComponentState, data["components"])
            self.selected_component = data["selected_component"]

    def initialize(self, supervoxel_ids):
        self.unused_supervoxels = set(supervoxel_ids)
        self.components = []
        self.selected_component = None

    def to_json(self):
        return {
            "unused_supervoxels": sorted(self.unused_supervoxels),
            "components": [x.to_json() for x in self.components],
            "selected_component": self.selected_component,
        }

    def save(self):
        if self.path is None:
            return
        tmp_path = self.path + ".tmp"
        with open(tmp_path, "w") as f:
            f.write(json.dumps(self.to_json()))
        os.rename(tmp_path, self.path)

    def make_new_component(self):
        c = ComponentState()
        c.supervoxels = self.unused_supervoxels
        self.unused_supervoxels = set()
        self.selected_component = len(self.components)
        self.components.append(c)

    def cycle_selected_component(self, amount):
        if len(self.components) == 0:
            return
        if self.selected_component is None:
            if amount > 0:
                self.selected_component = 0
            else:
                self.selected_component = len(self.components) - 1
        else:
            self.selected_component = (
                self.selected_component + amount + len(self.components)
            ) % len(self.components)

    def add_seed(self, supervoxel_id, position, inclusive):
        if self.selected_component is None:
            return
        c = self.components[self.selected_component]
        c.seeds[inclusive].append(
            dict(supervoxel_id=supervoxel_id, position=position, id=uuid.uuid4().hex)
        )


class CachedSplitResult:
    def __init__(self, state, graph, agglo_id):
        self.state = state
        self.graph = graph
        self.agglo_id = agglo_id
        self.reset()

    def reset(self):
        self.selected_component = None
        self.seeds = [[], []]
        self.supervoxels = set()
        self.split_result = None

    def update(self):
        selected_component = self.state.selected_component
        if selected_component is None:
            if self.selected_component is None:
                return False
            self.reset()
            return True
        component = self.state.components[selected_component]
        if selected_component == self.selected_component:
            if self.supervoxels == component.supervoxels:
                if self.seeds == component.seeds:
                    return False
        self.selected_component = self.state.selected_component
        self.seeds = copy.deepcopy(component.seeds)
        self.supervoxels = set(component.supervoxels)
        print("Recomputing split result")
        self.split_result = do_split(
            graph=self.graph,
            split_seeds=self.seeds,
            agglo_id=self.agglo_id,
            supervoxels=self.supervoxels,
        )
        print("Done recomputing split result")
        return True


class InteractiveSplitter:
    def __init__(self, graph, agglo_id, image_url, segmentation_url, state_path):
        self.graph = graph
        self.agglo_id = agglo_id
        self.image_url = image_url
        self.segmentation_url = segmentation_url
        self.state = InteractiveState(state_path)
        self.cached_split_result = CachedSplitResult(
            state=self.state, graph=self.graph, agglo_id=self.agglo_id
        )
        self.agglo_members = set(self.graph.get_agglo_members(agglo_id))

        if state_path is not None and os.path.exists(state_path):
            self.state.load()
        else:
            self.state.initialize(self.agglo_members)

        viewer = self.viewer = neuroglancer.Viewer()
        viewer.actions.add("inclusive-seed", self._add_inclusive_seed)
        viewer.actions.add("exclusive-seed", self._add_exclusive_seed)
        viewer.actions.add("next-component", self._next_component)
        viewer.actions.add("prev-component", self._prev_component)
        viewer.actions.add("new-component", self._make_new_component)
        viewer.actions.add("exclude-component", self._exclude_component)
        viewer.actions.add("exclude-all-but-component", self._exclude_all_but_component)

        key_bindings = [
            ["bracketleft", "prev-component"],
            ["bracketright", "next-component"],
            ["at:dblclick0", "exclude-component"],
            ["at:shift+mousedown2", "exclude-all-but-component"],
            ["at:control+mousedown0", "inclusive-seed"],
            ["at:shift+mousedown0", "exclusive-seed"],
            ["enter", "new-component"],
        ]

        with viewer.txn() as s:
            s.layers.append(
                name="image",
                layer=neuroglancer.ImageLayer(source=self.image_url),
            )
            s.layers.append(
                name="original",
                layer=neuroglancer.SegmentationLayer(
                    source=self.segmentation_url,
                    segments=self.agglo_members,
                ),
            )
            s.layers.append(
                name="unused",
                layer=neuroglancer.SegmentationLayer(
                    source=self.segmentation_url,
                ),
                visible=False,
            )
            s.layers.append(
                name="split-result",
                layer=neuroglancer.SegmentationLayer(
                    source=self.segmentation_url,
                    segments=self.agglo_members,
                ),
            )
            s.concurrent_downloads = 256
            self._update_state(s)

        with viewer.config_state.txn() as s:
            s.status_messages["help"] = "KEYS: " + " | ".join(
                f"{key}={command}" for key, command in key_bindings
            )
            for key, command in key_bindings:
                s.input_event_bindings.viewer[key] = command
                s.input_event_bindings.slice_view[key] = command
                s.input_event_bindings.perspective_view[key] = command
            self._update_config_state(s)

        viewer.shared_state.add_changed_callback(
            lambda: viewer.defer_callback(self._handle_state_changed)
        )

    def _add_inclusive_seed(self, s):
        self._add_seed(s, True)

    def _add_exclusive_seed(self, s):
        self._add_seed(s, False)

    def _exclude_component(self, s):
        if self.state.selected_component is None:
            return

        component = self.state.components[self.state.selected_component]
        supervoxel_id = self._get_mouse_supervoxel(s)

        if supervoxel_id is None:
            return

        self.cached_split_result.update()
        members = set(
            self.cached_split_result.split_result["cur_eqs"].members(supervoxel_id)
        )
        component.supervoxels = {x for x in component.supervoxels if x not in members}
        self.state.unused_supervoxels.update(members)
        self._update_view()

    def _exclude_all_but_component(self, s):
        if self.state.selected_component is None:
            return

        component = self.state.components[self.state.selected_component]
        supervoxel_id = self._get_mouse_supervoxel(s)

        if supervoxel_id is None:
            return

        self.cached_split_result.update()
        members = set(
            self.cached_split_result.split_result["cur_eqs"].members(supervoxel_id)
        )
        new_unused = {x for x in component.supervoxels if x not in members}
        component.supervoxels = members
        self.state.unused_supervoxels.update(new_unused)
        self._update_view()

    def _make_new_component(self, s):
        del s
        self.state.make_new_component()
        self._update_view()

    def _next_component(self, s):
        del s
        self.state.cycle_selected_component(1)
        self._update_view()

    def _prev_component(self, s):
        del s
        self.state.cycle_selected_component(-1)
        self._update_view()

    def _handle_state_changed(self):
        if self.state.selected_component is None:
            return
        seeds = _get_viewer_seeds(self.viewer.state)
        component = self.state.components[self.state.selected_component]
        if seeds == component.seeds:
            return
        component.seeds = seeds
        with self.viewer.txn() as s:
            self._update_state(s)

    def _get_mouse_supervoxel(self, s):
        supervoxel_id = s.selected_values.get("original")
        if supervoxel_id is not None:
            supervoxel_id = supervoxel_id.value
        if supervoxel_id is None:
            m = s.selected_values.get("split-result")
            if m is not None:
                m = m.value
            if m is not None:
                if isinstance(m, neuroglancer.SegmentIdMapEntry):
                    supervoxel_id = m.key
                else:
                    supervoxel_id = m
        if supervoxel_id is None or supervoxel_id == 0:
            return None
        return supervoxel_id

    def _add_seed(self, s, inclusive):
        supervoxel_id = self._get_mouse_supervoxel(s)
        mouse_voxel_coordinates = s.mouse_voxel_coordinates
        if mouse_voxel_coordinates is None or supervoxel_id is None:
            return
        position = tuple(int(x) for x in mouse_voxel_coordinates)
        self.state.add_seed(supervoxel_id, position, inclusive)
        self._update_view()

    def _update_view(self):
        with self.viewer.txn() as s:
            self._update_state(s)
        with self.viewer.config_state.txn() as s:
            self._update_config_state(s)

    def _update_config_state(self, s):
        if self.state.selected_component is None:
            msg = "[No component selected] %d unused supervoxels" % len(
                self.state.unused_supervoxels
            )
        else:
            selected_component = self.state.selected_component

            msg = (
                "[Component %d/%d] : %d supervoxels, %d connected components, %d unused"
                % (
                    selected_component,
                    len(self.state.components),
                    len(self.cached_split_result.supervoxels),
                    len(self.cached_split_result.split_result["cur_eqs"].sets()),
                    len(self.state.unused_supervoxels),
                )
            )
        s.status_messages["status"] = msg

    def _update_state(self, s):
        self.cached_split_result.update()
        self.state.save()
        _set_viewer_seeds(s, self.cached_split_result.seeds)

        s.layers["unused"].segments = self.state.unused_supervoxels
        s.layers["original"].segments = self.cached_split_result.supervoxels
        s.layers["split-result"].segments = self.cached_split_result.supervoxels
        split_result = self.cached_split_result.split_result
        if split_result is not None:
            self._show_split_result(
                s,
                cur_eqs=split_result["cur_eqs"],
            )
        s.layout = neuroglancer.row_layout(
            [
                neuroglancer.LayerGroupViewer(
                    layout="3d",
                    layers=[
                        "image",
                        "original",
                        "unused",
                        "inclusive-seeds",
                        "exclusive-seeds",
                    ],
                ),
                neuroglancer.LayerGroupViewer(
                    layout="3d",
                    layers=[
                        "image",
                        "split-result",
                        "inclusive-seeds",
                        "exclusive-seeds",
                    ],
                ),
            ]
        )

    def _show_split_result(self, s, cur_eqs):
        split_layer = s.layers["split-result"]
        split_layer.equivalences = cur_eqs
        split_layer.segments = {
            cur_eqs[x] for x in self.cached_split_result.supervoxels
        }


def run_batch(args, graph):
    for path in args.split_seeds:
        split_seeds = load_split_seeds(path)
        split_result = do_split(
            graph=graph, split_seeds=split_seeds, agglo_id=args.agglo_id
        )
        state = display_split_result(
            graph=graph,
            split_seeds=split_seeds,
            image_url=args.image_url,
            segmentation_url=args.segmentation_url,
            **split_result,
        )
        print(f'<p><a href="{neuroglancer.to_url(state)}">{path}</a></p>')


def run_interactive(args, graph):
    # Make splitter a global variable so that it is accessible from the
    # interactive `python -i` shell.
    global splitter

    neuroglancer.cli.handle_server_arguments(args)

    splitter = InteractiveSplitter(
        graph,
        agglo_id=args.agglo_id,
        image_url=args.image_url,
        segmentation_url=args.segmentation_url,
        state_path=args.state,
    )
    print(splitter.viewer)


def open_graph(path, agglo_id):
    # Check if graph_db is sharded
    graph_db = path
    m = re.match(r"(.*)@([0-9]+)((?:\..*)?)$", graph_db)
    if m is not None:
        num_shards = int(m.group(2))
        shard = agglo_id % num_shards
        graph_db = m.group(1) + ("-%05d-of-%05d" % (shard, num_shards)) + m.group(3)

    return AgglomerationGraph(sqlite3.connect(graph_db, check_same_thread=False))


if __name__ == "__main__":
    ap = argparse.ArgumentParser()

    ap.add_argument(
        "-v", "--verbose", action="store_true", help="Display verbose log messages."
    )

    common_ap = argparse.ArgumentParser(add_help=False)
    common_ap.add_argument(
        "--graph-db",
        required=True,
        help="Path to sqlite3 database specifying agglomeration graph",
    )
    common_ap.add_argument(
        "--image-url", required=True, help="Neuroglancer data source URL for image"
    )
    common_ap.add_argument(
        "--segmentation-url",
        required=True,
        help="Neuroglancer data source URL for segmentation",
    )

    sub_aps = ap.add_subparsers(help="command to run")
    interactive_ap = sub_aps.add_parser(
        "interactive",
        help="Interactively split an aglomerated component",
        parents=[common_ap],
    )

    batch_ap = sub_aps.add_parser(
        "batch", help="Split based on pre-specified seed files", parents=[common_ap]
    )

    interactive_ap.add_argument(
        "--agglo-id", type=int, required=True, help="Agglomerated component id to split"
    )
    interactive_ap.add_argument(
        "--split-seeds", help="Path to JSON file specifying split seeds"
    )
    interactive_ap.add_argument("--state", help="Path to JSON state file.")
    neuroglancer.cli.add_server_arguments(interactive_ap)

    interactive_ap.set_defaults(func=run_interactive)

    batch_ap.add_argument(
        "--split-seeds", nargs="+", help="Path to JSON file specifying split seeds"
    )
    batch_ap.add_argument(
        "--agglo-id", type=int, help="Agglomerated component id to split"
    )
    batch_ap.set_defaults(func=run_batch)

    args = ap.parse_args()

    graph = open_graph(args.graph_db, args.agglo_id)

    if args.verbose:
        logging.basicConfig(level=logging.INFO)

    args.func(args, graph)



# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import concurrent.futures
import json
import multiprocessing
import os
import re
import socket
import subprocess
import sys
import threading
import weakref

import numpy as np
import tornado.httpserver
import tornado.netutil
import tornado.platform.asyncio
import tornado.web

from . import async_util, local_volume, skeleton, static
from .json_utils import encode_json, json_encoder_default
from .random_token import make_random_token
from .trackable_state import ConcurrentModificationError

INFO_PATH_REGEX = r"^/neuroglancer/info/(?P<token>[^/]+)$"
SKELETON_INFO_PATH_REGEX = r"^/neuroglancer/skeletoninfo/(?P<token>[^/]+)$"

DATA_PATH_REGEX = r"^/neuroglancer/(?P<data_format>[^/]+)/(?P<token>[^/]+)/(?P<scale_key>[^/]+)/(?P<start>[0-9]+(?:,[0-9]+)*)/(?P<end>[0-9]+(?:,[0-9]+)*)$"

SKELETON_PATH_REGEX = r"^/neuroglancer/skeleton/(?P<key>[^/]+)/(?P<object_id>[0-9]+)$"

MESH_PATH_REGEX = r"^/neuroglancer/mesh/(?P<key>[^/]+)/(?P<object_id>[0-9]+)$"

STATIC_PATH_REGEX = (
    r"^/v/(?P<viewer_token>[^/]+)/(?P<path>(?:[@a-zA-Z0-9_\-][@a-zA-Z0-9_\-./]*)?)$"
)

ACTION_PATH_REGEX = r"^/action/(?P<viewer_token>[^/]+)$"

VOLUME_INFO_RESPONSE_PATH_REGEX = (
    r"^/volume_response/(?P<viewer_token>[^/]+)/(?P<request_id>[^/]+)/info$"
)

VOLUME_CHUNK_RESPONSE_PATH_REGEX = (
    r"^/volume_response/(?P<viewer_token>[^/]+)/(?P<request_id>[^/]+)/chunk$"
)

EVENTS_PATH_REGEX = r"^/events/(?P<viewer_token>[^/]+)$"

SET_STATE_PATH_REGEX = r"^/state/(?P<viewer_token>[^/]+)$"

CREDENTIALS_PATH_REGEX = r"^/credentials/(?P<viewer_token>[^/]+)$"

global_static_content_source = None

global_server_args = dict(bind_address="127.0.0.1", bind_port=0)

debug = False

_IS_GOOGLE_COLAB = "google.colab" in sys.modules


def _get_server_url(bind_address: str, port: int) -> str:
    if _IS_GOOGLE_COLAB:
        return _get_colab_server_url(port)
    return _get_regular_server_url(bind_address, port)


def _get_regular_server_url(bind_address: str, port: int) -> str:
    if bind_address == "0.0.0.0" or bind_address == "::":
        hostname = socket.getfqdn()
    else:
        hostname = bind_address
    return f"http://{hostname}:{port}"


def _get_colab_server_url(port: int) -> str:
    import google.colab.output

    return google.colab.output.eval_js(f"google.colab.kernel.proxyPort({port})")


class Server(async_util.BackgroundTornadoServer):
    def __init__(self, bind_address="127.0.0.1", bind_port=0, token=None):
        super().__init__(daemon=True)
        self.viewers = weakref.WeakValueDictionary()
        self._bind_address = bind_address
        self._bind_port = bind_port
        if token is None:
            token = make_random_token()
        self.token = token
        self.executor = concurrent.futures.ThreadPoolExecutor(
            max_workers=multiprocessing.cpu_count()
        )

    def _attempt_to_start_server(self):
        def log_function(handler):
            if debug:
                print(
                    "%d %s %.2fs"
                    % (
                        handler.get_status(),
                        handler.request.uri,
                        handler.request.request_time(),
                    )
                )

        app = self.app = tornado.web.Application(
            [
                (STATIC_PATH_REGEX, StaticPathHandler, dict(server=self)),
                (INFO_PATH_REGEX, VolumeInfoHandler, dict(server=self)),
                (SKELETON_INFO_PATH_REGEX, SkeletonInfoHandler, dict(server=self)),
                (DATA_PATH_REGEX, SubvolumeHandler, dict(server=self)),
                (SKELETON_PATH_REGEX, SkeletonHandler, dict(server=self)),
                (MESH_PATH_REGEX, MeshHandler, dict(server=self)),
                (ACTION_PATH_REGEX, ActionHandler, dict(server=self)),
                (
                    VOLUME_INFO_RESPONSE_PATH_REGEX,
                    VolumeInfoResponseHandler,
                    dict(server=self),
                ),
                (
                    VOLUME_CHUNK_RESPONSE_PATH_REGEX,
                    VolumeChunkResponseHandler,
                    dict(server=self),
                ),
                (EVENTS_PATH_REGEX, EventStreamHandler, dict(server=self)),
                (SET_STATE_PATH_REGEX, SetStateHandler, dict(server=self)),
                (CREDENTIALS_PATH_REGEX, CredentialsHandler, dict(server=self)),
            ],
            log_function=log_function,
        )
        self.http_server = tornado.httpserver.HTTPServer(
            app,
            # Allow very large requests to accommodate large screenshots.
            max_buffer_size=1024**3,
        )
        sockets = tornado.netutil.bind_sockets(
            port=self._bind_port, address=self._bind_address
        )
        self.http_server.add_sockets(sockets)
        actual_port = sockets[0].getsockname()[1]

        global global_static_content_source
        if global_static_content_source is None:
            global_static_content_source = static.get_default_static_content_source()
        self.port = actual_port
        self.server_url = _get_server_url(self._bind_address, actual_port)
        self.regular_server_url = _get_regular_server_url(
            self._bind_address, actual_port
        )
        self._credentials_manager = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()

    def get_volume(self, key):
        dot_index = key.find(".")
        if dot_index == -1:
            return None
        viewer_token = key[:dot_index]
        volume_token = key[dot_index + 1 :]
        viewer = self.viewers.get(viewer_token)
        if viewer is None:
            return None
        return viewer.volume_manager.volumes.get(volume_token)


class BaseRequestHandler(tornado.web.RequestHandler):
    def initialize(self, server):
        self.server = server


class StaticPathHandler(BaseRequestHandler):
    def get(self, viewer_token, path):
        if (
            viewer_token != self.server.token
            and viewer_token not in self.server.viewers
        ):
            self.send_error(404)
            return
        try:
            query = self.request.query
            if query:
                query = f"?{query}"
            data, content_type = global_static_content_source.get(path, query)
        except ValueError as e:
            self.send_error(404, message=e.args[0])
            return
        self.set_header("Content-type", content_type)
        self.finish(data)


class ActionHandler(BaseRequestHandler):
    def post(self, viewer_token):
        viewer = self.server.viewers.get(viewer_token)
        if viewer is None:
            self.send_error(404)
            return
        action = json.loads(self.request.body)
        self.server.loop.call_soon(
            viewer.actions.invoke, action["action"], action["state"]
        )
        self.finish("")


class VolumeInfoResponseHandler(BaseRequestHandler):
    def post(self, viewer_token, request_id):
        viewer = self.server.viewers.get(viewer_token)
        if viewer is None:
            self.send_error(404)
            return

        info = json.loads(self.request.body)
        self.server.loop.call_soon(viewer._handle_volume_info_reply, request_id, info)
        self.finish("")


class VolumeChunkResponseHandler(BaseRequestHandler):
    def post(self, viewer_token, request_id):
        viewer = self.server.viewers.get(viewer_token)
        if viewer is None:
            self.send_error(404)
            return

        params = json.loads(self.get_argument("p"))
        data = self.request.body
        self.server.loop.call_soon(
            viewer._handle_volume_chunk_reply, request_id, params, data
        )
        self.finish("")


class EventStreamStateWatcher:
    def __init__(self, key: str, client_id: str, state, last_generation: str, wake_up):
        self.key = key
        self.state = state
        self.last_generation = last_generation
        self._wake_up = wake_up
        self._client_id = client_id
        state.add_changed_callback(wake_up)

    def unregister(self):
        self.state.remove_changed_callback(self._wake_up)

    def maybe_send_update(self, handler):
        raw_state, generation = self.state.raw_state_and_generation
        if generation == self.last_generation:
            return False
        if generation.startswith(self._client_id + "/"):
            return False
        self.last_generation = generation
        msg = {"k": self.key, "s": raw_state, "g": generation}
        handler.write(f"data: {encode_json(msg)}\n\n")
        if debug:
            print(f"data: {encode_json(msg)}\n\n")
        return True


class EventStreamHandler(BaseRequestHandler):
    async def get(self, viewer_token: str):
        viewer = self.server.viewers.get(viewer_token)
        if viewer is None:
            self.send_error(404)
            return
        self.set_header("content-type", "text/event-stream")
        self.set_header("cache-control", "no-cache")
        must_flush = True
        wake_event = asyncio.Event()
        self._wake_up = lambda: self.server.loop.call_soon_threadsafe(wake_event.set)
        client_id = self.get_query_argument("c")
        if client_id is None:
            raise tornado.web.HTTPError(400, "missing client_id")
        self._closed = False

        watchers = []

        def watch(key: str, state):
            last_generation = self.get_query_argument(f"g{key}")
            if last_generation is None:
                raise tornado.web.HTTPError(400, f"missing g{key}")
            watchers.append(
                EventStreamStateWatcher(
                    key=key,
                    state=state,
                    client_id=client_id,
                    wake_up=self._wake_up,
                    last_generation=last_generation,
                )
            )

        watch("c", viewer.config_state)
        if hasattr(viewer, "shared_state"):
            watch("s", viewer.shared_state)

        try:
            while True:
                wake_event.clear()
                if self._closed:
                    break
                try:
                    sent = False
                    for watcher in watchers:
                        sent = watcher.maybe_send_update(self) or sent
                    if sent:
                        self.flush()
                        if _IS_GOOGLE_COLAB:
                            # The proxy used by colab buffers the entire
                            # response.  Therefore we must end the response for
                            # the client to receive it.
                            break
                    elif must_flush:
                        self.flush()
                except tornado.iostream.StreamClosedError:
                    break
                await wake_event.wait()
                must_flush = False
        finally:
            for watcher in watchers:
                watcher.unregister()

    def on_connection_close(self):
        if debug:
            print("connection closed")
        super().on_connection_close()
        self._wake_up()
        self._closed = True


class SetStateHandler(BaseRequestHandler):
    def post(self, viewer_token: str):
        viewer = self.server.viewers.get(viewer_token)
        if viewer is None:
            self.send_error(404)
            return
        msg = json.loads(self.request.body)
        prev_generation = msg["pg"]
        generation = msg["g"]
        state = msg["s"]
        client_id = msg["c"]
        try:
            new_generation = viewer.set_state(
                state, f"{client_id}/{generation}", existing_generation=prev_generation
            )
            self.set_header("Content-type", "application/json")
            self.finish(json.dumps({"g": new_generation}))
        except ConcurrentModificationError:
            self.set_status(412)
            self.finish("")


class CredentialsHandler(BaseRequestHandler):
    async def post(self, viewer_token: str):
        viewer = self.server.viewers.get(viewer_token)
        if viewer is None:
            self.send_error(404)
            return
        if not viewer.allow_credentials:
            self.send_error(403)
            return
        if self.server._credentials_manager is None:
            from .default_credentials_manager import default_credentials_manager

            self.server._credentials_manager = default_credentials_manager
        msg = json.loads(self.request.body)
        invalid = msg.get("invalid")
        provider = self.server._credentials_manager.get(
            msg["key"], msg.get("parameters")
        )
        if provider is None:
            self.send_error(400)
            return
        try:
            credentials = await asyncio.wrap_future(provider.get(invalid))
            self.set_header("Content-type", "application/json")
            self.finish(json.dumps(credentials))
        except Exception:
            import traceback

            traceback.print_exc()
            self.send_error(401)


class VolumeInfoHandler(BaseRequestHandler):
    def get(self, token):
        vol = self.server.get_volume(token)
        if vol is None or not isinstance(vol, local_volume.LocalVolume):
            self.send_error(404)
            return
        self.finish(json.dumps(vol.info(), default=json_encoder_default).encode())


class SkeletonInfoHandler(BaseRequestHandler):
    def get(self, token):
        vol = self.server.get_volume(token)
        if vol is None or not isinstance(vol, skeleton.SkeletonSource):
            self.send_error(404)
            return
        self.finish(json.dumps(vol.info(), default=json_encoder_default).encode())


class SubvolumeHandler(BaseRequestHandler):
    async def get(self, data_format, token, scale_key, start, end):
        start_pos = np.array(start.split(","), dtype=np.int64)
        end_pos = np.array(end.split(","), dtype=np.int64)
        vol = self.server.get_volume(token)
        if vol is None or not isinstance(vol, local_volume.LocalVolume):
            self.send_error(404)
            return

        try:
            data, content_type = await asyncio.wrap_future(
                self.server.executor.submit(
                    vol.get_encoded_subvolume,
                    data_format=data_format,
                    start=start_pos,
                    end=end_pos,
                    scale_key=scale_key,
                )
            )
        except ValueError as e:
            self.send_error(400, message=e.args[0])
            return
        self.set_header("Content-type", content_type)
        self.finish(data)


class MeshHandler(BaseRequestHandler):
    async def get(self, key, object_id):
        object_id = int(object_id)
        vol = self.server.get_volume(key)
        if vol is None or not isinstance(vol, local_volume.LocalVolume):
            self.send_error(404)
            return

        try:
            encoded_mesh = await asyncio.wrap_future(
                self.server.executor.submit(vol.get_object_mesh, object_id)
            )
        except local_volume.MeshImplementationNotAvailable:
            self.send_error(501, message="Mesh implementation not available")
            return
        except local_volume.MeshesNotSupportedForVolume:
            self.send_error(405, message="Meshes not supported for volume")
            return
        except local_volume.InvalidObjectIdForMesh:
            self.send_error(404, message="Mesh not available for specified object id")
            return
        except ValueError as e:
            self.send_error(400, message=e.args[0])
            return

        self.set_header("Content-type", "application/octet-stream")
        self.finish(encoded_mesh)


class SkeletonHandler(BaseRequestHandler):
    async def get(self, key, object_id):
        object_id = int(object_id)
        vol = self.server.get_volume(key)
        if vol is None or not isinstance(vol, skeleton.SkeletonSource):
            self.send_error(404)
            return

        def get_encoded_skeleton(skeletons, object_id):
            skeleton = skeletons.get_skeleton(object_id)
            if skeleton is None:
                return None
            return skeleton.encode(skeletons)

        try:
            encoded_skeleton = await asyncio.wrap_future(
                self.server.executor.submit(get_encoded_skeleton, vol, object_id)
            )
        except Exception as e:
            self.send_error(500, message=e.args[0])
            return
        if encoded_skeleton is None:
            self.send_error(
                404, message="Skeleton not available for specified object id"
            )
            return
        self.set_header("Content-type", "application/octet-stream")
        self.finish(encoded_skeleton)


global_server = None
_global_server_lock = threading.Lock()


def _get_server_token():
    with _global_server_lock:
        if global_server is not None:
            return global_server.token
        token = make_random_token()
        global_server_args.update(token=token)
        return token


def set_static_content_source(*args, **kwargs):
    """Specifies the Neuroglancer client hosted by the Python integration webserver.

    Group:
      server
    """
    global global_static_content_source
    global_static_content_source = static.get_static_content_source(*args, **kwargs)


def set_dev_server_content_source():
    """Builds the Neuroglancer client from source on demand for the Python integration.

    Group:
      Server
    """
    static_content_url = None
    root_dir = os.path.join(os.path.dirname(__file__), "..", "..")
    build_process = subprocess.Popen(
        [
            "npm",
            "run",
            "dev-server-python",
            "--",
            "--port=0",
            "--no-lint",
            "--no-typecheck",
        ],
        cwd=root_dir,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        encoding="utf-8",
    )

    try:
        future = concurrent.futures.Future()

        def thread_func(f):
            url = None
            for line in f:
                print(f"[dev-server] {line.rstrip()}")
                if url is None:
                    m = re.search(r"http://[^\s,]+", line)
                    if m is not None:
                        url = m.group(0)
                        future.set_result(url)
            if url is None:
                future.set_result(None)

        thread = threading.Thread(target=thread_func, args=(build_process.stdout,))
        thread.daemon = True
        thread.start()

        static_content_url = future.result(timeout=10)
    except:
        build_process.terminate()
        raise

    set_static_content_source(url=static_content_url)


def set_server_bind_address(bind_address=None, bind_port=0):
    """Sets the bind adress and port for the Python integration webserver.

    If the server is already running, this does not have any effect until the server is
    restarted.

    Group:
      Server

    """
    if bind_address is None:
        bind_address = "127.0.0.1"
    with _global_server_lock:
        global_server_args.update(bind_address=bind_address, bind_port=bind_port)


def is_server_running() -> bool:
    """Returns ``True`` if the Python integration webserver is running.

    Group:
      Server
    """
    with _global_server_lock:
        return global_server is not None


def stop() -> None:
    """Stop the server, invalidating any viewer URLs.

    This allows any previously-referenced data arrays to be garbage collected if there are no other
    references to them.

    Group:
      server
    """
    global global_server
    with _global_server_lock:
        server = global_server
        global_server = None
    if server is not None:
        server.stop()


def get_server_url() -> str:
    if global_server is None:
        raise ValueError("Server is not started")
    return global_server.server_url


def start() -> None:
    """Starts the Python integration webserver if not already started.

    Group:
      Server
    """
    global global_server
    with _global_server_lock:
        if global_server is not None:
            return

        # Workaround https://bugs.python.org/issue37373
        # https://www.tornadoweb.org/en/stable/index.html#installation
        if sys.platform == "win32" and sys.version_info >= (3, 8):
            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
        global_server = Server(**global_server_args)


def register_viewer(viewer):
    start()
    global_server.viewers[viewer.token] = viewer


def defer_callback(callback, *args, **kwargs):
    """Register `callback` to run in the server event loop thread.

    Group:
      Server
    """
    start()
    global_server.loop.call_soon_threadsafe(lambda: callback(*args, **kwargs))



# @license
# Copyright 2016 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

import numpy as np


def downsample_with_averaging(array, factor):
    """Downsample x by factor using averaging.

    @return: The downsampled array, of the same type as x.
    """
    factor = tuple(factor)
    output_shape = tuple(int(math.ceil(s / f)) for s, f in zip(array.shape, factor))
    temp = np.zeros(output_shape, dtype=np.float32)
    counts = np.zeros(output_shape, int)
    for offset in np.ndindex(factor):
        part = array[tuple(np.s_[o::f] for o, f in zip(offset, factor))]
        indexing_expr = tuple(np.s_[:s] for s in part.shape)
        temp[indexing_expr] += part
        counts[indexing_expr] += 1
    return np.asarray(temp / counts, dtype=array.dtype)


def downsample_with_striding(array, factor):
    """Downsample x by factor using striding.

    @return: The downsampled array, of the same type as x.
    """
    return array[tuple(np.s_[::f] for f in factor)]



# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import collections
import copy
from collections.abc import ItemsView, Iterator, KeysView


class EquivalenceMap:
    """Union-find data structure.

    Group:
      viewer-state-segments
    """

    supports_readonly = True

    def __init__(self, existing=None, _readonly=False):
        """Create a new empty union-find structure."""
        if isinstance(existing, EquivalenceMap):
            self._weights = existing._weights.copy()
            self._parents = existing._parents.copy()
            self._prev_next = existing._prev_next.copy()
            self._min_values = existing._min_values.copy()
        else:
            self._weights = {}
            self._parents = {}
            self._prev_next = {}
            self._min_values = {}
            self._readonly = False
            if existing is not None:
                if isinstance(existing, dict):
                    existing = existing.items()
                for group in existing:
                    self.union(*group)
        self._readonly = _readonly

    def _get_representative(self, obj: int) -> int:
        """Finds and returns the root of the set containing the specified element."""

        if obj not in self._parents:
            self._parents[obj] = obj
            self._weights[obj] = 1
            self._prev_next[obj] = [obj, obj]
            self._min_values[obj] = obj
            return obj

        path = [obj]
        root = self._parents[obj]
        while root != path[-1]:
            path.append(root)
            root = self._parents[root]

        # compress the path and return
        for ancestor in path:
            self._parents[ancestor] = root
        return root

    def __getitem__(self, obj: int) -> int:
        """Returns the minimum element in the set containing the specified element."""
        if obj not in self._parents:
            return obj
        return self._min_values[self._get_representative(obj)]

    def __iter__(self) -> Iterator[int]:
        """Iterates over all elements known to this equivalence map."""
        return iter(self._parents)

    def items(self) -> ItemsView[int, int]:
        return self._parents.items()

    def keys(self) -> KeysView[int]:
        return self._parents.keys()

    def clear(self):
        self._weights.clear()
        self._parents.clear()
        self._prev_next.clear()
        self._min_values.clear()

    def union(self, *elements: int) -> int | None:
        """Unions the equivalence classes containing the specified elements."""
        if self._readonly:
            raise AttributeError

        if len(elements) == 0:
            return None
        if len(elements) == 1:
            return self[elements[0]]
        for a, b in zip(elements[:-1], elements[1:]):
            result = self._union_pair(a, b)
        return result

    def _union_pair(self, a: int, b: int) -> int:
        a = self._get_representative(a)
        b = self._get_representative(b)
        if a == b:
            return self._min_values[a]
        if self._weights[a] < self._weights[b]:
            a, b = (b, a)

        self._min_values[a] = min(self._min_values[a], self._min_values[b])

        a_links_new = self._prev_next[a]
        a_links_old = tuple(a_links_new)
        b_links_new = self._prev_next[b]
        b_links_old = tuple(b_links_new)

        # We want to splice b's list at the end of a's list

        # Splice beginning of b's list into end of a's list
        b_links_new[0] = a_links_old[0]
        self._prev_next[a_links_old[0]][1] = b

        # Splice end of b's list into end of a's list

        # last element of a's list is set to last element of b's list
        a_links_new[0] = b_links_old[0]
        # fix next pointer for last element of b's list
        self._prev_next[b_links_old[0]][1] = a
        self._weights[a] += self._weights[b]
        self._parents[b] = a
        return self._min_values[a]

    def members(self, element: int) -> Iterator[int]:
        """Yields the members of the equivalence class containing the specified element."""
        if element not in self._parents:
            yield element
            return
        cur_x = element
        while True:
            yield cur_x
            cur_x = self._prev_next[cur_x][1]
            if cur_x == element:
                break

    def sets(self) -> frozenset[frozenset[int]]:
        """Returns the equivalence classes as a set of sets."""
        sets: dict[int, set[int]] = collections.defaultdict(set)
        for x in self._parents:
            sets[self[x]].add(x)
        return frozenset(frozenset(v) for v in sets.values())

    def to_json(self) -> list[list[int]]:
        """Returns the equivalence classes a sorted list of sorted lists."""
        sets = self.sets()
        return sorted(sorted(x) for x in sets)

    def __copy__(self) -> "EquivalenceMap":
        """Does not preserve _readonly attribute."""
        return EquivalenceMap(self)

    def __deepcopy__(self, memo) -> "EquivalenceMap":
        """Does not preserve _readonly attribute."""
        result = EquivalenceMap()
        result._parents = copy.deepcopy(self._parents, memo)
        result._weights = copy.deepcopy(self._weights, memo)
        result._prev_next = copy.deepcopy(self._prev_next, memo)
        result._min_values = copy.deepcopy(self._min_values, memo)
        return result

    def copy(self) -> "EquivalenceMap":
        """Returns a copy of the equivalence map."""
        return EquivalenceMap(self)

    def delete_set(self, element: int):
        """Removes the equivalence class containing the specified element."""
        if element not in self._parents:
            return
        members = list(self.members(element))
        for v in members:
            del self._parents[v]
            del self._weights[v]
            del self._prev_next[v]
            del self._min_values[v]

    def isolate_element(self, element: int):
        """Isolates the specified element from its equivalence class."""
        members = list(self.members(element))
        self.delete_set(element)
        self.union(*(v for v in members if v != element))



"""Tools to convert between affine representations."""

# externals
import numpy as np

# internals
from .layout import layout_matrix
from .lie import lie_to_matrix, matrix_to_lie


class Affine:
    """Base class for all affine transformations"""

    def xras(self):
        """First column of a compact Vox2World matrix (MGH convention)"""
        return self.xyzcras()[..., 0]

    def yras(self):
        """Second column of a compact Vox2World matrix (MGH convention)"""
        return self.xyzcras()[..., 1]

    def zras(self):
        """Third column of a compact Vox2World matrix (MGH convention)"""
        return self.xyzcras()[..., 2]

    def cras(self):
        """Last column (translation) of a compact Vox2World matrix
         (MGH convention)"""
        return self.xyzcras()[..., -1]

    def xyzcras(self):
        """Compact form of a Vox2World matrix (MGH convention)"""
        return self.compact()

    def sform(self):
        """Alias for the compact matrix form of a Vox2World (NIfTI convention)"""
        return self.homogenous()[..., :-1, :]

    def compact(self):
        """Compact form of a homogeneous affine matrix (last row dropped)"""
        return self.homogenous()[..., :-1, :]

    def homogenous(self):
        """Homogeneous matrix form."""
        matrix = self.compact()
        new_shape = list(matrix.shape)
        new_shape[-2] += 1
        new_matrix = np.zeros_like(matrix, shape=new_shape)
        new_matrix[..., :-1, :] = matrix
        new_matrix[..., -1, -1] = 1
        return new_matrix

    def affine(self):
        """Alias for the homogeneous matrix form (NiBabel convention)"""
        return self.homogenous()

    def translation(self):
        """Translation component"""
        return self.compact()[..., -1]

    def linear(self):
        """Linear component"""
        return self.compact()[..., :-1]

    def rotation(self):
        """Rotation component (in the least squares sense)"""
        basis = 'GL+'
        matrix = self.linear().homogeneous()
        lie = matrix_to_lie(matrix, basis)
        rot = lie[:3]
        return lie_to_matrix(rot, basis, ndim=3)

    def quaternion(self):
        """Rotation component encoded with quaternion"""
        return RotationMatrix(self.rotation).quaternion()

    def qform(self):
        """Rotation component encoded with quaternion (NIfTI convention)"""
        return self.quaternion()

    def axisangle(self):
        """Rotation encoded by an axis of rotation and an angle"""
        return Quaternion(self.quaternion).axisangle()

    def axis(self):
        """Rotation encoded by an axis of rotation and an angle"""
        return Quaternion(self.quaternion).axis()

    def angle(self):
        """Rotation encoded by an axis of rotation and an angle"""
        return Quaternion(self.quaternion).angle()

    def voxel_size(self):
        """Voxel size of a Voxel2World"""
        return self.linear().square().sum(-1).sqrt()


class HomogeneousAffineMatrix(Affine):
    """ (D+1)x(D+1) homogeneous affine matrix"""

    def __init__(self, matrix):
        matrix = np.asanyarray(matrix)
        if matrix.shape[-1] != matrix.shape[-2]:
            raise ValueError('Expected square matrix')
        self.matrix = matrix

    def homogenous(self):
        return self.matrix


class CompactAffineMatrix(Affine):
    """ Dx(D+1) compact affine matrix"""

    def __init__(self, matrix):
        matrix = np.asanyarray(matrix)
        if matrix.shape[-1] - 1 != matrix.shape[-2]:
            raise ValueError('Expected compact matrix')
        self.matrix = matrix

    def compact(self):
        return self.matrix


class XYZC(Affine):
    """Affine matrix stored using independent columns"""

    def __init__(self, x, y, z, c):
        self.x = x
        self.y = y
        self.z = z
        self.c = c

    def xras(self):
        return self.x

    def yras(self):
        return self.y

    def zras(self):
        return self.z

    def cras(self):
        return self.c

    def compact(self):
        return np.stack([self.x, self.y, self.z, self.c], axis=-1)


class Linear(Affine):
    """Affine subgroup (no translation)"""

    def compact(self):
        matrix = self.linear()
        new_shape = list(matrix.shape)
        new_shape[-1] += 1
        new_matrix = np.zeros_like(matrix, shape=new_shape)
        new_matrix[..., :-1] = matrix
        return new_matrix


class Rotation(Linear):
    """Linear subgroup (orthogonal)"""

    def linear(self):
        return self.rotation()


class RotationMatrix(Rotation):
    """Rotation encoded using a DxD orthogonal matrix."""

    def __init__(self, matrix):
        matrix = np.asanyarray(matrix)
        if matrix.shape[-1] != matrix.shape[-2]:
            raise ValueError('Expected a square matrix')
        self.matrix = matrix

    def rotation(self):
        return self.matrix

    def quaternion(self):
        def vpos(mat, trace):
            r = trace.add(1).sqrt()
            s = r.reciprocal().div(2)
            i = (mat[..., 2, 1] - mat[..., 1, 2]) * s
            j = (mat[..., 0, 2] - mat[..., 2, 0]) * s
            k = (mat[..., 1, 0] - mat[..., 0, 1]) * s
            w = r.div(2)
            return np.stack([i, j, k, w], -1)

        def vneg(mat):
            r = (1 + mat[..., 0, 0] - mat[..., 1, 1] - mat[..., 2, 2]).sqrt()
            s = r.reciprocal().div(2)
            w = (mat[..., 2, 1] - mat[..., 1, 2]) * s
            i = r.div(2)
            j = (mat[..., 0, 1] - mat[..., 1, 0]) * s
            k = (mat[..., 2, 0] - mat[..., 0, 2]) * s
            return np.stack([i, j, k, w], -1)

        trace = self.matrix.trace()
        mask = trace > 0
        return np.where(mask, vpos(self.matrix, trace), vneg(self.matrix))


class Quaternion(Rotation):
    """Rotation stored in quanternion form"""

    def __init__(self, *args):
        """

        Parameters
        ----------
        Either
            quat : (..., 4) tensor
        Or
            orientation : (..., 3) tensor
            attitude : (...) tensor
        Or
            i, j, k, r : tensors

        """
        if len(args) == 1:
            ijkr = np.asanyarray(args[0])
            i = ijkr[..., 0]
            j = ijkr[..., 1]
            k = ijkr[..., 2]
            r = ijkr[..., 3]
        elif len(args) == 2:
            ijk, r = args
            i = ijk[..., 0]
            j = ijk[..., 1]
            k = ijk[..., 2]
        elif len(args) == 4:
            i, j, k, r = args
        else:
            raise ValueError('Expected 1, 2 or 4 arguments')
        self.i = i
        self.j = j
        self.k = k
        self.r = r

    def quaternion(self):
        return self.i, self.j, self.k, self.r

    def rotation(self):
        i = self.i
        j = self.j
        k = self.k
        r = self.r
        matrix = [[1-2*(j**2 + k**2), 2*(i*j - k*r), 2*(i*k + j*r)],
                  [2*(i*j + k*r), 1 - 2*(i**2 + k**2), 2*(j*k - i*r)],
                  [2*(i*k - j*r), 2*(j*k + i*r), 1 - 2*(i**2 + j**2)]]
        matrix = np.asanyarray(matrix)
        matrix = np.moveaxis(matrix, [0, 1], [-2, -1])
        return matrix

    def axisangle(self):
        axis = np.stack([self.i, self.j, self.k], axis=-1)
        nrm = (axis*axis).sum(-1, keepdims=True).sqrt()
        axis = axis / nrm
        nrm = nrm[..., 0]
        angle = 2*np.atan2(nrm, self.r)[..., None]
        return np.concatenate([axis, angle], axis=-1)

    def axis(self):
        axis = np.stack([self.i, self.j, self.k], axis=-1)
        axis = axis / (axis*axis).sum(-1, keepdims=True).sqrt()
        return axis

    def angle(self):
        axis = np.stack([self.i, self.j, self.k], axis=-1)
        axis = (axis*axis).sum(-1).sqrt()
        angle = 2*np.atan2(axis, self.r)
        return angle


class AxisAngle(Rotation):
    """Rotation stored using an axis of rotation and an angle."""

    def __init__(self, axis, angle):
        self.ax = axis
        self.theta = angle

    def quaternion(self):
        return Quaternion(self.ax * self.theta.div(2).sin(),
                          self.theta.div(2).cos()).quaternion()

    def axis(self):
        return self.ax

    def angle(self):
        return self.theta

    def axisangle(self):
        return np.concatenate([self.axis, self.angle[..., None]], -1)


class Orientation(Affine):
    """Orientation matrix that contains only an anisotropic scaling,
    flips and permutations, and a translation.

    This is the simplest possible voxel-to-world matrix for
    images of real-world objects.
    """

    def __init__(self, shift, scale, orientation='RAS'):
        self.shift = shift
        self.scale = scale
        self.orientation = orientation

    def affine(self):
        aff = layout_matrix(self.orientation, self.scale)
        aff[:-1, -1] = self.shift
        return aff



"""Color tags that can be used in a terminal prompt."""


def rgb_to_ansi256(r: int, g: int, b: int) -> int:
    """Convert RGB [0-255] to ANSI256 code."""
    # we use the extended greyscale palette here, with the exception of
    # black and white. normal palette only has 4 greyscale shades.
    r, g, b = int(round(r)), int(round(g)), int(round(b))
    if (r == g and g == b):
        if (r < 8):
            return 16
        if (r > 248):
            return 231
        return round(((r - 8) / 247) * 24) + 232

    return int(
        16 +
        (36 * round(r / 255 * 5)) +
        (6 * round(g / 255 * 5)) +
        round(b / 255 * 5)
    )


class tcolors:
    """Namespace that maps colors to integers for use in terminals."""

    (endc, bold, faint, italic, underline, blink) = range(6)

    class fg:
        """Foreground colors."""

        (black, red, green, yellow, blue, magenta, cyan, white) \
            = range(30, 38)

        class bright:
            """Bright foreground colors."""

            (black, red, green, yellow, blue, magenta, cyan, white) \
                = range(90, 98)

    class bg:
        """Background colors."""

        (black, red, green, yellow, blue, magenta, cyan, white) \
            = range(40, 48)

        class bright:
            """Bright background colors."""

            (black, red, green, yellow, blue, magenta, cyan, white) \
                = range(100, 108)

    header = fg.bright.magenta
    okblue = fg.bright.blue
    okcyan = fg.bright.cyan
    okgreen = fg.bright.green
    warning = fg.bright.yellow
    fail = fg.bright.red


class bcolors:
    """Namespace that maps colors to tokens for use in bash terminals."""


class icolors:
    """Namespace that maps colors to tokens for use in bash input prompts."""


class bformat:
    """Namespace for formatting functions, for use in bash terminals."""


class iformat:
    """Namespace for formatting functions, for use in bash input prompts."""


# To add color in the input prompt, all delimiter must be wrapped
# in this. See:
# https://bugs.python.org/issue12972
# https://stackoverflow.com/questions/9468435/
L = '\001'
R = '\002'


def _copy_attrs(dst: type, src: type, fn: callable) -> None:
    """Copy attributes (and classes) from source class to dest class."""
    for key, val in src.__dict__.items():
        if key.startswith("_"):
            continue
        if isinstance(val, type):
            setattr(dst, key, type(key, (), {}))
            _copy_attrs(getattr(dst, key), getattr(src, key), fn)
        else:
            setattr(dst, key, fn(getattr(src, key)))


def _make_format(color: str, endc: str) -> classmethod:
    return classmethod(lambda _, x: f"{color}{x}{endc}")


_copy_attrs(bcolors, tcolors, lambda x: f'\033[{x}m')
_copy_attrs(icolors, bcolors, lambda x: L + x + R)
_copy_attrs(bformat, bcolors, lambda x: _make_format(x, bcolors.endc))
_copy_attrs(iformat, icolors, lambda x: _make_format(x, icolors.endc))


tcolors.fg.rgb256 = staticmethod(rgb_to_ansi256)
bcolors.fg.rgb256 = staticmethod(
    lambda *a: f'\033[38;5;{tcolors.fg.rgb256(*a)}m'
)
bformat.fg.rgb256 = staticmethod(
    lambda *a: (lambda x: f"{bcolors.fg.rgb256(*a)}{x}{bcolors.endc}")
)
iformat.fg.rgb256 = staticmethod(
    lambda *a: (lambda x: f"{icolors.fg.rgb256(*a)}{x}{icolors.endc}")
)

tcolors.bg.rgb256 = staticmethod(rgb_to_ansi256)
bcolors.bg.rgb256 = staticmethod(
    lambda *a: f'\033[48;5;{tcolors.bg.rgb256(*a)}m'
)
bformat.bg.rgb256 = staticmethod(
    lambda *a: (lambda x: f"{bcolors.bg.rgb256(*a)}{x}{bcolors.endc}")
)
iformat.bg.rgb256 = staticmethod(
    lambda *a: (lambda x: f"{icolors.bg.rgb256(*a)}{x}{icolors.endc}")
)



# @license
# Copyright 2016 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib.resources
import os
import posixpath
import re

mime_type_map = {
    ".css": "text/css",
    ".js": "application/javascript",
    ".html": "text/html",
    ".map": "application/json",
    ".wasm": "application/wasm",
}


def guess_mime_type_from_path(path):
    return mime_type_map.get(posixpath.splitext(path)[1], "application/octet-stream")


class StaticContentSource:
    def get(self, name, query):
        if name == "":
            name = "index.html"
        return self.get_content(name), guess_mime_type_from_path(name)

    def get_content(self, name):
        raise NotImplementedError


class ImportlibResourcesContentSource(StaticContentSource):
    def get_content(self, name):
        if not re.match(r"^[a-z0-9][a-zA-Z0-9_\-\.]*\.(?:js|map|css|wasm|html)$", name):
            raise ValueError("Invalid static resource name: %r" % name)
        path = importlib.resources.files(__name__).joinpath("client", name)
        if path.is_file():
            return path.read_bytes()
        raise ValueError(
            'Static resources not built.  Run: "npm run build-python" or use an alternative static content source.'
        )


PkgResourcesContentSource = ImportlibResourcesContentSource


class HttpSource(StaticContentSource):
    def __init__(self, url):
        self.url = url

    def get(self, name, query):
        import requests

        print("http", repr(name), repr(query))

        full_url = posixpath.join(self.url, name)
        r = requests.get(full_url + query)
        if r.status_code >= 200 and r.status_code < 300:
            return r.content, (
                r.headers.get("content-type") or guess_mime_type_from_path(name)
            )
        raise ValueError(f"Failed to retrieve {full_url!r}: {r.reason}")


class FileSource(StaticContentSource):
    def __init__(self, path, file_open=None):
        self.file_path = path
        self.file_open = file_open or open

    def get_content(self, name):
        full_path = os.path.join(self.file_path, name)
        try:
            with self.file_open(full_path, "rb") as f:
                return f.read()
        except Exception as e:
            raise ValueError(f"Failed to read local path {full_path!r}: {e}")


def get_default_static_content_source():
    return ImportlibResourcesContentSource()


def get_static_content_source(source=None, url=None, path=None, file_open=None):
    if source is not None:
        return source
    elif url is not None:
        return HttpSource(url)
    elif path is not None:
        return FileSource(path=path, file_open=file_open)
    else:
        return get_default_static_content_source()



"""A neuroglancer scene with a programmatic interface."""
# stdlib
import logging
from os import PathLike
from typing import Iterator

# externals
import numpy as np

# internals
from ngtools.datasources import (
    LayerDataSources,
    MeshDataSource,
    SkeletonDataSource,
    VolumeDataSource,
)
from ngtools.local.tracts import TractDataSource, TractSkeleton
from ngtools.opener import parse_protocols
from ngtools.optionals import try_import_as
from ngtools.shaders import shaders
from ngtools.utils import Wraps

# optionals
ng = try_import_as('neuroglancer', fallback='ngtools._nglite')

LOG = logging.getLogger(__name__)

URILike = str | PathLike
SourceType = ng.LayerDataSource | ng.LocalVolume | ng.skeleton.SkeletonSource


_LocalType = (ng.local_volume.LocalVolume, ng.skeleton.SkeletonSource)
_LocalLike = ng.local_volume.LocalVolume | ng.skeleton.SkeletonSource
_DataSourceLike = (
    str |
    ng.local_volume.LocalVolume |
    ng.skeleton.SkeletonSource |
    ng.LayerDataSource
)
_DataSourcesLike = (
    _DataSourceLike |
    list[_DataSourceLike | dict | None] |
    (ng.LayerDataSources | dict | None)
)
_LayerLike = ng.Layer | dict | None
_LayerArg = (
    _DataSourceLike |
    list[_DataSourceLike | dict | None] |
    ng.LayerDataSources |
    (ng.Layer | dict | None)
)


def _get_url(arg: _LayerArg, **kwargs) -> str | _LocalLike | None:
    if kwargs.get("source", None):
        return _get_url(kwargs.pop("source"), **kwargs)
    if kwargs.get("url", None):
        return _get_url(kwargs.pop("url"), **kwargs)
    if isinstance(arg, (str, PathLike)):
        return str(arg)
    if isinstance(arg, _LocalType):
        if hasattr(arg, "_url"):
            return arg._url
        return arg
    if hasattr(arg, "source"):
        return _get_url(arg.source)
    if hasattr(arg, "url"):
        return _get_url(arg.url)
    if isinstance(arg, dict):
        if arg.get("source", None):
            return _get_url(arg.get("source", arg))
        if arg.get("url", None):
            return _get_url(arg.get("url", arg))
    if hasattr(arg, "__iter__"):
        for arg1 in arg:
            url = _get_url(arg1, **kwargs)
            if url is not None:
                return url
    print("None")
    return None


def _get_source(arg: _LayerArg, **kwargs) -> _DataSourcesLike:
    if kwargs.get("source", None):
        return _get_url(kwargs.pop("source"), **kwargs)
    if isinstance(arg, (str, PathLike)):
        return str(arg)
    if isinstance(arg, _LocalType):
        return arg
    if hasattr(arg, "source"):
        return _get_url(arg.source)
    if hasattr(arg, "url"):
        return arg
    if isinstance(arg, dict):
        if arg.get("source", None):
            return _get_url(arg.get("source", arg))
        if "url" in arg:
            return arg
    if hasattr(arg, "__iter__"):
        return arg
    if isinstance(arg, (ng.LayerDataSource, ng.LayerDataSources)):
        return arg
    return None


class LayerFactory(type):
    """Metaclass for layers."""

    def __call__(cls, arg: _LayerArg = None, *args, **kwargs) -> "Layer":
        """
        Parameters
        ----------
        arg : ng.Layer or [list of] (str | dict | None)
        """
        arg_json = arg.to_json() if hasattr(arg, "to_json") else arg
        LOG.debug(f"LayerFactory({cls.__name__}, {arg_json})")

        # Only use the factory if it is not called from a subclass
        if cls is not Layer:
            LOG.debug(f"LayerFactory - defer to {cls.__name__}")
            return super().__call__(arg, *args, **kwargs)

        # Special case: pointAnnotation do not have `source`
        if (
            isinstance(arg, ng.PointAnnotationLayer) or
            kwargs.get("type", "") == "pointAnnotation" or
            (
                isinstance(arg, dict) and
                arg.get("type", "") == "pointAnnotation"
            )
        ):
            LOG.debug("LayerFactory - PointAnnotationLayer")
            return PointAnnotationLayer(arg, *args, **kwargs)

        # Check whether the main input is source-like object
        if not (
            arg is None or
            isinstance(arg, ng.Layer) or
            (isinstance(arg, dict) and "url" not in arg)
        ):
            if "source" in kwargs:
                raise ValueError(
                    "source-like object found in both positional "
                    "and keyword arguments."
                )
            kwargs["source"] = arg
            arg = None

        # Switch based on layer protocol
        url = _get_url(arg, **kwargs)
        LOG.debug(f"LayerFactory - url: {url}")

        GuessedLayer = None
        if url and isinstance(url, str):
            if url == "local://annotations":
                if isinstance(arg, ng.LocalAnnotationLayer):
                    odim = arg.source[0].transform.output_dimensions
                    return LocalAnnotationLayer(odim, arg, *args, **kwargs)
                else:
                    return AnnotationLayer(arg, *args, **kwargs)

            layer_type = parse_protocols(url).layer
            LOG.debug(f"LayerFactory - hint: {layer_type}")
            GuessedLayer = {
                "image": ImageLayer,
                "volume": ImageLayer,
                "segmentation": SegmentationLayer,
                "labels": SegmentationLayer,
                "mesh": MeshLayer,
                "surface": MeshLayer,
                "skeleton": SkeletonLayer,
                "tracts": TractLayer,
                "annotation": AnnotationLayer,
            }.get(layer_type, None)
            if GuessedLayer:
                return GuessedLayer(arg, *args, **kwargs)

        # Switch based on type keyword
        if hasattr(arg, "type"):
            layer_type = arg.type
        elif isinstance(arg, dict):
            layer_type = arg.get("type", None)
        else:
            layer_type = None
        LOG.debug(f"LayerFactory - json hint: {layer_type}")

        GuessedLayer = {
            "image": ImageLayer,
            "segmentation": SegmentationLayer,
            "mesh": SingleMeshLayer,
            "annotation": AnnotationLayer,
        }.get(layer_type, None)
        if GuessedLayer:
            return GuessedLayer(arg, *args, **kwargs)

        # Switch based on data source type
        LOG.debug("LayerFactory - build sources")
        sources = LayerDataSources(kwargs.get("source", None))
        kwargs["source"] = sources

        for source in sources:
            LOG.debug(f"LayerFactory - source: {str(type(source))}")
            if isinstance(source, VolumeDataSource):
                LOG.debug("LayerFactory - guess ImageLayer")
                GuessedLayer = ImageLayer
            elif isinstance(source, SkeletonDataSource):
                LOG.debug("LayerFactory - guess SkeletonLayer")
                GuessedLayer = SkeletonLayer
            elif isinstance(source, MeshDataSource):
                LOG.debug("LayerFactory - guess MeshLayer")
                GuessedLayer = MeshLayer
            if GuessedLayer:
                # kwargs["source"] = sources
                return GuessedLayer(*args, **kwargs)

        # Fallback
        LOG.debug("LayerFactory - fallback to simple Layer")
        return super().__call__(arg, *args, **kwargs)


class _SourceMixin:

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        # Re-build data sources so that parameters for which
        # ngtools and neuroglancer use different defaults values
        # get set up according to ngtools conventions.
        LOG.debug("Layer - recompute data sources")
        self.source = LayerDataSources(self.source)

    def __get_source__(self) -> LayerDataSources:
        return LayerDataSources(self._wrapped.source)


class Layer(Wraps(ng.Layer), metaclass=LayerFactory):
    """Base class for all layers.

    See `ImageLayer`, `SegmentationLayer`, `MeshLayer` for
    keyword-parameters specific to these layers types.

    Parameters
    ----------
    json_data : dict | Layer | LayerDataSource | str
        Either a Layer (or its JSON representation),
        or a data source (or its JSON representation).

    Other Parameters
    ----------------
    type : {"image", "segmentation", "mesh", "annotation", "pointAnnotation"}
        Layer type -- guessed from source by default.
    source : [list of] LayerDataSource | str
        Data source. Except if layer is a `PointAnnotationLayer`.
    layer_dimensions : ng.CoordinateSpace
        Local dimensions specific to the layer.
    layer_position : ng.CoordinateSpace
        Position within the local dimensions.
    local_velocity : ng.DimensionPlaybackVelocity
        Speed for playback along local dimensions.
    tab : str
    panels : list[ng.LayerSidePanelState]
    pick : bool
    tool_bindings : dict[str, Tool | str]
        Key binding to tools, specific to this layer.
    tool : Tool | str
        Active tool (or tool name).
    """  # noqa: E501

    def __init__(self, *args, **kwargs) -> None:
        LOG.debug("Layer.__init__")
        super().__init__(*args, **kwargs)
        if hasattr(self, "source"):
            # Re-build data sources so that parameters for which
            # ngtools and neuroglancer use different defaults values
            # get set up according to ngtools conventions.
            LOG.debug("Layer - recompute data sources")
            self.source = LayerDataSources(self.source)

    def apply_transform(self, *args: ng.CoordinateSpaceTransform) -> "Layer":
        """Apply an additional transform in model space."""
        for source in getattr(self, "source", []):
            getattr(source, "apply_transform", (lambda *_: None))(*args)
        return self


"""
NOTE: Valid tools are
```
{
    # AnnotationLayer
    "annotatePoint": ng.PlacePointTool,
    "annotateLine": ng.PlaceLineTool,
    "annotateBoundingBox": ng.PlaceBoundingBoxTool,
    "annotateSphere": ng.PlaceEllipsoidTool,
    # ImageLayer
    "blend": ng.BlendTool,
    "opacity": ng.OpacityTool,
    "volumeRendering": ng.VolumeRenderingTool,
    "volumeRenderingGain": ng.VolumeRenderingGainTool,
    "volumeRenderingDepthSamples": ng.VolumeRenderingDepthSamplesTool,
    "crossSectionRenderScale": ng.CrossSectionRenderScaleTool,
    # SegmentationLayer
    "selectedAlpha": ng.SelectedAlphaTool,
    "notSelectedAlpha": ng.NotSelectedAlphaTool,
    "objectAlpha": ng.ObjectAlphaTool,
    "hideSegmentZero": ng.HideSegmentZeroTool,
    "hoverHighlight": ng.HoverHighlightTool,
    "baseSegmentColoring": ng.BaseSegmentColoringTool,
    "ignoreNullVisibleSet": ng.IgnoreNullVisibleSetTool,
    "colorSeed": ng.ColorSeedTool,
    "segmentDefaultColor": ng.SegmentDefaultColorTool,
    # MeshLayer
    "meshRenderScale": ng.MeshRenderScaleTool,
    "meshSilhouetteRendering": ng.MeshSilhouetteRenderingTool,
    "saturation": ng.SaturationTool,
    # SkeletonLayer
    "skeletonRendering.mode2d": ng.SkeletonRenderingMode2dTool,
    "skeletonRendering.mode3d": ng.SkeletonRenderingMode3dTool,
    "skeletonRendering.lineWidth2d": ng.SkeletonRenderingLineWidth2dTool,
    "skeletonRendering.lineWidth3d": ng.SkeletonRenderingLineWidth3dTool,
    # All
    "shaderControl": ng.ShaderControlTool,
    "dimension": ng.DimensionTool,
    # SegmentEdit
    "mergeSegments: ng.MergeSegmentsTool,
    "splitSegments: ng.SplitSegmentsTool,
    "selectSegments: ng.SelectSegmentsTool,
```
"""


class ImageLayer(_SourceMixin, Wraps(ng.ImageLayer), Layer):
    """Image layer.

    See `Layer` for keyword parameters common to all layers.

    Other Parameters
    ----------------
    source : [list of] LayerDataSource | str
    shader : str
        Default shader:
        ```glsl
        #uicontrol invlerp normalized
            void main() {
            emitGrayscale(normalized());
        }
        ```
    shader_controls : dict[str, float | dict | ng.InvlerpParameters | ng.TransferFunctionParameters]
        - InvlerpParameters:
            range: tuple[float, float]
            window: tuple[float, float]
            channel: list[int]
        - TransferFunctionParameters:
            window: tuple[float, float]
            channel: list[int]
            controlPoints : list[list[float]]
            defaultColor : str
    opacity : float, default=0.5
    blend : {"default", "additive"}, default="default"
    volume_rendering_mode : bool | {"off", "on", "max", "min"}, default=False
    volume_rendering_gain : float, default=0
    volume_rendering_depth_samples : float, default=64
    cross_section_render_scale : float, default=1
    """  # noqa: E501

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        if self.shader == "None":
            self.shader = shaders.default

        if self.source and not self.shaderControls:
            source = self.source[0]
            try:
                mn, q0, q1, mx = source.quantiles([0.0, 0.01, 0.99, 1.0])
                self.shaderControls = {
                    "normalized": {
                        "range": np.stack([q0, q1]).tolist(),
                        "window": np.stack([mn, mx]).tolist(),
                    }
                }
            except Exception as e:
                # The source may not be accessible, fail gracefully
                url = getattr(source, "url", None)
                url = getattr(source, "local_url", url)
                LOG.error(f"Could not compute quantiles for {url}: {e}")
                pass


class SegmentationLayer(_SourceMixin, Wraps(ng.SegmentationLayer), Layer):
    """Segmentation layer.

    See `Layer` for keyword parameters common to all layers.

    Other Parameters
    ----------------
    source : [list of] LayerDataSource | str
    starred_segments : StarredSegments | dict[int, bool]
    equivalences : ng.EquivalenceMap | list[list[int]]
    hide_segment_zero : bool, default=True
    hover_highlight : bool, default=True
    base_segment_coloring : bool, default=False
    selected_alpha : float, default=0.5
    not_selected_alpha : float, default=0
    object_alpha : float, default=1
    saturation : float, default=1
    ignore_null_visible_set : bool, default=True
    skeleton_rendering : ng.SkeletonRenderingOptions | dict
        shader : str, optional
        shader_controls : dict[str, float | dict | ng.InvlerpParameters | ng.TransferFunctionParameters]
            - InvlerpParameters:
                range: tuple[float, float]
                window: tuple[float, float]
                channel: list[int]
            - TransferFunctionParameters:
                window: tuple[float, float]
                channel: list[int]
                controlPoints : list[list[float]]
                defaultColor : str
        mode2d : {"lines", "lines_and_points"}, optional
        line_width2d : float, default=2
        mode3d: {"lines", "lines_and_points"}, optional
        line_width3d : float, default=3
    color_seed : int, default=0
    cross_section_render_scale : float, default=1
    mesh_render_scale : float, default=10
    mesh_silhouette_rendering : float, default=0
    segment_query : str, optional
    segment_colors : dict[uint64, str]
    segment_default_color : str, optional
    linked_segmentation_group : str, optional
    linked_segmentation_color_group : str | False
    """  # noqa: E501

    ...


class SkeletonLayerFactory(LayerFactory):
    """Factory for skeletons."""

    def __call__(  # noqa: D102
        cls, arg: _LayerArg = None, *args, **kwargs
    ) -> "SkeletonLayer":
        json_arg = arg.to_json() if hasattr(arg, "to_json") else arg
        LOG.debug(f"SkeletonLayerFactory({cls.__name__}, {json_arg})")

        # Only use the factory if it is not called from a subclass
        if cls is not SkeletonLayer:
            LOG.debug(f"SkeletonLayer - defer to {cls.__name__}")
            return super().__call__(arg, *args, **kwargs)

        # Check whether the main input is source-like object
        if not (
            arg is None or
            isinstance(arg, ng.Layer) or
            (isinstance(arg, dict) and "url" not in arg)
        ):
            if "source" in kwargs:
                raise ValueError(
                    "source-like object found in both positional "
                    "and keyword arguments."
                )
            kwargs["source"] = arg
            arg = None

        # Switch based on layer protocol
        url = _get_url(arg, **kwargs)
        LOG.debug(f"SkeletonLayerFactory - url: {url}")

        GuessedLayer = None
        if url:
            if isinstance(url, str):
                layer_type = parse_protocols(url).layer
                LOG.debug(f"SkeletonLayerFactory - hint: {layer_type}")
                GuessedLayer = {
                    "tracts": TractLayer,
                    "skeleton": SkeletonLayer,
                }.get(layer_type, None)
            elif isinstance(url, TractSkeleton):
                GuessedLayer = TractLayer
            elif isinstance(url, ng.skeleton.SkeletonSource):
                GuessedLayer = SkeletonLayer
            if GuessedLayer:
                if GuessedLayer is SkeletonLayer:
                    return super().__call__(arg, *args, **kwargs)
                else:
                    return GuessedLayer(arg, *args, **kwargs)

        # Switch based on data source type
        sources = LayerDataSources(kwargs.get("source", None))
        for source in sources:
            LOG.debug(f"SkeletonLayerFactory - source: {str(type(source))}")
            if isinstance(source, TractDataSource):
                LOG.debug("SkeletonLayerFactory - guessed TractLayer")
                GuessedLayer = TractLayer
            elif isinstance(source, SkeletonDataSource):
                LOG.debug("SkeletonLayerFactory - guessed SkeletonLayer")
                GuessedLayer = SkeletonLayer
            if GuessedLayer:
                kwargs["source"] = sources
                if GuessedLayer is SkeletonLayer:
                    return super().__call__(arg, *args, **kwargs)
                else:
                    return GuessedLayer(arg, *args, **kwargs)

        msg = f"Cannot guess layer type for {source}"
        LOG.error(f"SkeletonLayerFactory - {msg}")
        raise ValueError(msg)


class SkeletonLayer(SegmentationLayer, metaclass=SkeletonLayerFactory):
    """Skeleton layer."""

    shader = SegmentationLayer.skeleton_shader

    @property
    def shader_controls(self):  # noqa: ANN201, D102
        return self.skeleton_rendering.shader_controls

    @shader_controls.setter
    def shader_controls(self, value) -> None:  # noqa: ANN001
        self.skeleton_rendering.shader_controls = value

    @property
    def mode2d(self):  # noqa: ANN201, D102
        return self.skeleton_rendering.mode2d

    @mode2d.setter
    def mode2d(self, value) -> None:  # noqa: ANN001
        self.skeleton_rendering.mode2d = value

    @property
    def mode3d(self):  # noqa: ANN201, D102
        return self.skeleton_rendering.mode3d

    @mode3d.setter
    def mode3d(self, value) -> None:  # noqa: ANN001
        self.skeleton_rendering.mode3d = value

    @property
    def line_width2d(self):  # noqa: ANN201, D102
        return self.skeleton_rendering.line_width2d

    @line_width2d.setter
    def line_width2d(self, value) -> None:  # noqa: ANN001
        self.skeleton_rendering.line_width2d = value

    @property
    def line_width3d(self):  # noqa: ANN201, D102
        return self.skeleton_rendering.line_width3d

    @line_width3d.setter
    def line_width3d(self, value) -> None:  # noqa: ANN001
        self.skeleton_rendering.line_width3d = value


class TractLayer(SkeletonLayer):
    """Tract layer."""

    def __init__(self, arg: _LayerArg = None, *args, **kwargs) -> None:
        json_arg = arg.to_json() if hasattr(arg, "to_json") else arg
        LOG.debug(f"TractLayer({json_arg})")

        # already a TractLayer -> no deep copy
        if isinstance(arg, TractLayer) and not (args or kwargs):
            LOG.debug(f"TractLayer - defer to {self.__wrapped_class__}")
            super().__init__(arg)
            return

        # get source parameters
        if not (
            arg is None or
            isinstance(arg, ng.Layer) or
            (isinstance(arg, dict) and "url" not in arg)
        ):
            if "source" in kwargs:
                raise ValueError(
                    "source-like object found in both positional "
                    "and keyword arguments."
                )
            kwargs["source"] = arg
            arg = None
            LOG.debug(f"TractLayer - source: {kwargs['source']}")

        # split layer / source keywords
        ksrc = {}
        for key in kwargs:
            if not hasattr(TractLayer, key):
                ksrc[key] = kwargs.pop(key)

        # build source
        LOG.debug("TractLayer - build source...")
        source = kwargs["source"]
        if isinstance(source, (ng.LayerDataSources, list, tuple)):
            if len(source) != 1:
                raise ValueError(f"too many sources: {len(source)}")
            source = source[0]
        kwargs["source"] = TractDataSource(source, **ksrc)
        LOG.debug(f"TractLayer - build source: {kwargs['source'].to_json()}")

        LOG.debug(f"TractLayer - defer to {self.__wrapped_class__}")
        super().__init__(arg, *args, **kwargs)

        if self.skeleton_rendering.shader is None:
            self.skeleton_rendering.shader = shaders.skeleton.orientation
        if self.skeleton_rendering.mode2d is None:
            self.skeleton_rendering.mode2d = "lines"
        if self.skeleton_rendering.lineWidth2d == 2:
            self.skeleton_rendering.lineWidth2d = 0.01
        if self.selected_alpha is None:
            self.selected_alpha = 1
        if self.not_selected_alpha is None:
            self.not_selected_alpha = 0
        if not self.segments:
            self.segments = [1]

        # if "nbtracts" not in self.shader_controls:
        #     max_tracts = len(source.url.tractfile.streamlines)
        #     self.shader_controls["nbtracts"] = {
        #         "min": 0,
        #         "max": max_tracts,
        #         "default": source.url.max_tracts,
        #     }


class MeshLayer(Layer):
    """Common ancestors to all mesh layers.

    See also: `SingleMeshLayer`, `MultiscaleMeshLayer`
    """

    ...


class SingleMeshLayer(_SourceMixin, Wraps(ng.SingleMeshLayer), MeshLayer):
    """
    Single mesh layer.

    See `Layer` for keyword parameters common to all layers.

    See also: `MeshLayer`, `MultiscaleMeshLayer`.

    Other Parameters
    ----------------
    source : [list of] LayerDataSource | str
    vertex_attribute_sources : list[str]
    shader : str, optional
        Default shader:
        ```glsl
        void main() {
            emitGray();
        }
        ```
    vertex_attribute_names : list[str]
    """

    ...


class MultiscaleMeshLayer(SegmentationLayer):
    """Multiscale mesh layer.

    See also: `MeshLayer`, `SingleMeshLayer`.

    Other Parameters
    ----------------
    render_scale : float, default=10
    silhouette_rendering : float, default=0
    """

    render_scale = SegmentationLayer.mesh_render_scale
    silhouette_rendering = SegmentationLayer.mesh_silhouette_rendering


class AnnotationLayer(_SourceMixin, Wraps(ng.AnnotationLayer), Layer):
    """Annotation layer.

    See `Layer` for keyword parameters common to all layers.

    Other Parameters
    ----------------
    source : [list of] LayerDataSource | str
    annotations : list[AnnotationBase | dict]
        Valid annotations:
        ```
        {
            "point": PointAnnotation,
            "line": LineAnnotation,
            "axis_aligned_bounding_box": AxisAlignedBoundingBoxAnnotation,
            "ellipsoid": EllipsoidAnnotation,
        }
        ```
    annotation_properties : list[AnnotationPropertySpec | dict]
        id : str
        type : str
        description : str
        default : float | str
        enum_values : list[int | str]
        enum_labels : list[str]
    annotation_relationships : list[str]
    linked_segmentation_layer : dict[str, str]
    filter_by_segmentation : list[str]
    ignore_null_segment_filter : bool, default=True
    shader : str, optional
        Default shader:
        ```glsl
        void main() {
            setColor(defaultColor());
        }
        ```
    shader_controls : dict[str, float | dict | ng.InvlerpParameters | ng.TransferFunctionParameters]
        - InvlerpParameters:
            range: tuple[float, float]
            window: tuple[float, float]
            channel: list[int]
        - TransferFunctionParameters:
            window: tuple[float, float]
            channel: list[int]
            controlPoints : list[list[float]]
            defaultColor : str
    """  # noqa: E501

    ...


class LocalAnnotationLayer(Wraps(ng.LocalAnnotationLayer), AnnotationLayer):
    """TODO."""

    ...


class PointAnnotationLayer(Wraps(ng.PointAnnotationLayer), Layer):
    """Point annotation layer.

    Parameters
    ----------
    points : list[tuple[float, float, float]]
    """

    ...


class ManagedLayer(Wraps(ng.ManagedLayer), _SourceMixin):
    """Named layer."""

    def __get_layer__(self) -> Layer:
        return Layer(self._wrapped.layer)


class Layers(Wraps(ng.Layers)):
    """List of named layers."""

    def __getitem__(self, *args, **kwargs) -> ManagedLayer:
        return ManagedLayer(super().__getitem__(*args, **kwargs))

    def __iter__(self) -> Iterator[ManagedLayer]:
        for layer in super().__iter__():
            yield ManagedLayer(layer)



# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Surface transforms."""
import pathlib
import warnings
import numpy as np
import nibabel as nb
from scipy import sparse
from scipy.spatial import KDTree
from scipy.spatial.distance import cdist
from .base import (
    SurfaceMesh
)


try:
    import h5py
    _H5_AVAILABLE = True
except ImportError:
    _H5_AVAILABLE = False


class SurfaceTransformBase():
    """Generic surface transformation class"""

    def __init__(self, reference, moving, spherical=False):
        """Instantiate a generic surface transform."""
        if spherical:
            if not reference.check_sphere():
                raise ValueError("reference was not spherical")
            if not moving.check_sphere():
                raise ValueError("moving was not spherical")
            reference.set_radius()
            moving.set_radius()
        self._reference = reference
        self._moving = moving

    def __eq__(self, other):
        ref_coords_eq = np.all(self.reference._coords == other.reference._coords)
        ref_tris_eq = np.all(self.reference._triangles == other.reference._triangles)
        mov_coords_eq = np.all(self.moving._coords == other.moving._coords)
        mov_tris_eq = np.all(self.moving._triangles == other.moving._triangles)
        return ref_coords_eq & ref_tris_eq & mov_coords_eq & mov_tris_eq

    def __invert__(self):
        return self.__class__(self._moving, self._reference)

    @property
    def reference(self):
        return self._reference

    @reference.setter
    def reference(self, surface):
        self._reference = SurfaceMesh(surface)

    @property
    def moving(self):
        return self._moving

    @moving.setter
    def moving(self, surface):
        self._moving = SurfaceMesh(surface)

    @classmethod
    def from_filename(cls, reference_path, moving_path):
        """Create an Surface Index Transformation from a pair of surfaces with corresponding
         vertices."""
        reference = SurfaceMesh(nb.load(reference_path))
        moving = SurfaceMesh(nb.load(moving_path))
        return cls(reference, moving)


class SurfaceCoordinateTransform(SurfaceTransformBase):
    """Represents surface transformations in which the indices correspond and the coordinates
     differ. This could be two surfaces representing difference structures from the same
     hemisphere, like white matter and pial, or it could be a sphere and a deformed sphere that
     moves those coordinates to a different location."""

    __slots__ = ("_reference", "_moving")

    def __init__(self, reference, moving):
        """Instantiate a transform between two surfaces with corresponding vertices.
         Parameters
        ----------
        reference: surface
            Surface with the starting coordinates for each index.
        moving: surface
            Surface with the destination coordinates for each index.
        """

        super().__init__(reference=SurfaceMesh(reference), moving=SurfaceMesh(moving))
        if np.all(self._reference._triangles != self._moving._triangles):
            raise ValueError("Both surfaces for an index transform must have corresponding"
                             " vertices.")

    def map(self, x, inverse=False):
        if not inverse:
            source = self.reference
            dest = self.moving
        else:
            source = self.moving
            dest = self.reference

        s_tree = KDTree(source._coords)
        dists, matches = s_tree.query(x)
        if not np.allclose(dists, 0):
            raise NotImplementedError("Mapping on surfaces not implemented for coordinates that"
                                      " aren't vertices")
        return dest._coords[matches]

    def __add__(self, other):
        if isinstance(other, SurfaceCoordinateTransform):
            return self.__class__(self.reference, other.moving)
        raise NotImplementedError

    def _to_hdf5(self, x5_root):
        """Write transform to HDF5 file."""
        if not _H5_AVAILABLE:
            raise ImportError("h5py")
        triangles = x5_root.create_group("Triangles")
        coords = x5_root.create_group("Coordinates")
        coords.create_dataset("0", data=self.reference._coords)
        coords.create_dataset("1", data=self.moving._coords)
        triangles.create_dataset("0", data=self.reference._triangles)
        xform = x5_root.create_group("Transform")
        xform.attrs["Type"] = "SurfaceCoordinateTransform"
        reference = xform.create_group("Reference")
        reference['Coordinates'] = h5py.SoftLink('/0/Coordinates/0')
        reference['Triangles'] = h5py.SoftLink('/0/Triangles/0')
        moving = xform.create_group("Moving")
        moving['Coordinates'] = h5py.SoftLink('/0/Coordinates/1')
        moving['Triangles'] = h5py.SoftLink('/0/Triangles/0')

    def to_filename(self, filename, fmt=None):
        """Store the transform."""
        if fmt is None:
            fmt = "npz" if filename.endswith(".npz") else "X5"

        if fmt == "npz":
            raise NotImplementedError
            # sparse.save_npz(filename, self.mat)
            # return filename

        if not _H5_AVAILABLE:
            raise ImportError("h5py")

        with h5py.File(filename, "w") as out_file:
            out_file.attrs["Format"] = "X5"
            out_file.attrs["Version"] = np.uint16(1)
            root = out_file.create_group("/0")
            self._to_hdf5(root)

        return filename

    @classmethod
    def from_filename(cls, filename=None, reference_path=None, moving_path=None,
                      fmt=None):
        """Load transform from file."""
        if filename is None:
            if reference_path is None or moving_path is None:
                raise ValueError("You must pass either a X5 file or a pair of reference and moving"
                                 " surfaces.")
            return cls(SurfaceMesh(nb.load(reference_path)),
                       SurfaceMesh(nb.load(moving_path)))

        if fmt is None:
            try:
                fmt = "npz" if filename.endswith(".npz") else "X5"
            except AttributeError:
                fmt = "npz" if filename.as_posix().endswith(".npz") else "X5"

        if fmt == "npz":
            raise NotImplementedError
            # return cls(sparse.load_npz(filename))

        if fmt != "X5":
            raise ValueError("Only npz and X5 formats are supported.")

        if not _H5_AVAILABLE:
            raise ImportError("h5py")

        with h5py.File(filename, "r") as f:
            assert f.attrs["Format"] == "X5"
            xform = f["/0/Transform"]
            reference = SurfaceMesh.from_arrays(
                xform['Reference']['Coordinates'],
                xform['Reference']['Triangles']
            )

            moving = SurfaceMesh.from_arrays(
                xform['Moving']['Coordinates'],
                xform['Moving']['Triangles']
            )
        return cls(reference, moving)


class SurfaceResampler(SurfaceTransformBase):
    """
    Represents transformations in which the coordinate space remains the same
    and the indices change.
    To achieve surface project-unproject functionality:
        sphere_in as the reference
        sphere_project_to as the moving
    Then apply the transformation to sphere_unproject_from
    """

    __slots__ = ("_reference", "_moving", "mat", 'interpolation_method')

    def __init__(self, reference, moving, interpolation_method='barycentric', mat=None):
        """Initialize the resampling.

        Parameters
        ----------
        reference: spherical surface of the reference space.
            Output will have number of indices equal to the number of indicies in this surface.
            Both reference and moving should be in the same coordinate space.
        moving: spherical surface that will be resampled.
            Both reference and moving should be in the same coordinate space.
        mat : array-like, shape (nv1, nv2)
            Sparse matrix representing the transform.
        interpolation_method : str
            Only barycentric is currently implemented
        """

        super().__init__(SurfaceMesh(reference), SurfaceMesh(moving), spherical=True)

        self.reference.set_radius()
        self.moving.set_radius()
        if interpolation_method not in ['barycentric']:
            raise NotImplementedError(f"{interpolation_method} is not implemented.")
        self.interpolation_method = interpolation_method

        # TODO: should we deal with the case where reference and moving are the same?

        # we're calculating the interpolation in the init so that we can ensure
        # that it only has to be calculated once and will always be saved with the
        # transform
        if mat is None:
            self.__calculate_mat()
            m_tree = KDTree(self.moving._coords)
            _, kmr_closest = m_tree.query(self.reference._coords, k=10)

            # invert the triangles to generate a lookup table from vertices to triangle index
            tri_lut = {}
            for i, idxs in enumerate(self.moving._triangles):
                for x in idxs:
                    if x not in tri_lut:
                        tri_lut[x] = [i]
                    else:
                        tri_lut[x].append(i)

            # calculate the barycentric interpolation weights
            bc_weights = []
            enclosing = []
            for point, kmrv in zip(self.reference._coords, kmr_closest):
                close_tris = _find_close_tris(kmrv, tri_lut, self.moving)
                ww, ee = _find_weights(point, close_tris, m_tree)
                bc_weights.append(ww)
                enclosing.append(ee)

            # build sparse matrix
            # commenting out code for barycentric nearest neighbor
            # bary_nearest = []
            mat = sparse.lil_array((self.reference._npoints, self.moving._npoints))
            for s_ix, dd in enumerate(bc_weights):
                for k, v in dd.items():
                    mat[s_ix, k] = v
                # bary_nearest.append(np.array(list(dd.keys()))[np.array(list(dd.values())).argmax()])
            # bary_nearest = np.array(bary_nearest)
            # transpose so that number of out vertices is columns
            self.mat = sparse.csr_array(mat.T)
        else:
            if isinstance(mat, sparse.csr_array):
                self.mat = mat
            else:
                self.mat = sparse.csr_array(mat)
            # validate shape of the provided matrix
            if (mat.shape[0] != moving._npoints) or (mat.shape[1] != reference._npoints):
                msg = "Shape of provided mat does not match expectations based on " \
                      "dimensions of moving and reference. \n"
                if mat.shape[0] != moving._npoints:
                    msg += f" mat has {mat.shape[0]} rows but moving has {moving._npoints} " \
                           f"vertices. \n"
                if mat.shape[1] != reference._npoints:
                    msg += f" mat has {mat.shape[1]} columns but reference has" \
                           f" {reference._npoints} vertices."
                raise ValueError(msg)

    def __calculate_mat(self):
        m_tree = KDTree(self.moving._coords)
        _, kmr_closest = m_tree.query(self.reference._coords, k=10)

        # invert the triangles to generate a lookup table from vertices to triangle index
        tri_lut = {}
        for i, idxs in enumerate(self.moving._triangles):
            for x in idxs:
                if x not in tri_lut:
                    tri_lut[x] = [i]
                else:
                    tri_lut[x].append(i)

        # calculate the barycentric interpolation weights
        bc_weights = []
        enclosing = []
        for point, kmrv in zip(self.reference._coords, kmr_closest):
            close_tris = _find_close_tris(kmrv, tri_lut, self.moving)
            ww, ee = _find_weights(point, close_tris, m_tree)
            bc_weights.append(ww)
            enclosing.append(ee)

        # build sparse matrix
        # commenting out code for barycentric nearest neighbor
        # bary_nearest = []
        mat = sparse.lil_array((self.reference._npoints, self.moving._npoints))
        for s_ix, dd in enumerate(bc_weights):
            for k, v in dd.items():
                mat[s_ix, k] = v
            # bary_nearest.append(
            #   np.array(list(dd.keys()))[np.array(list(dd.values())).argmax()]
            # )
        # bary_nearest = np.array(bary_nearest)
        # transpose so that number of out vertices is columns
        self.mat = sparse.csr_array(mat.T)

    def map(self, x):
        return x

    def __add__(self, other):
        if (isinstance(other, SurfaceResampler)
                and (other.interpolation_method == self.interpolation_method)):
            return self.__class__(
                self.reference,
                other.moving,
                interpolation_method=self.interpolation_method
            )
        raise NotImplementedError

    def __invert__(self):
        return self.__class__(
            self.moving,
            self.reference,
            interpolation_method=self.interpolation_method
        )

    @SurfaceTransformBase.reference.setter
    def reference(self, surface):
        raise ValueError("Don't modify the reference of an existing resampling."
                         "Create a new one instead.")

    @SurfaceTransformBase.moving.setter
    def moving(self, surface):
        raise ValueError("Don't modify the moving of an existing resampling."
                         "Create a new one instead.")

    def apply(self, x, inverse=False, normalize="element"):
        """Apply the transform to surface data.

        Parameters
        ----------
        x : array-like, shape (..., nv1), or SurfaceMesh
            Data to transform or SurfaceMesh to resample
        inverse : bool, default=False
            Whether to apply the inverse transform. If True, ``x`` has shape
            (..., nv2), and the output will have shape (..., nv1).
        normalize : {"element", "sum", "none"}, default="element"
            Normalization strategy. If "element", the scale of each value in
            the output is comparable to each value of the input. If "sum", the
            sum of the output is comparable to the sum of the input. If
            "none", no normalization is applied.

        Returns
        -------
        y : array-like, shape (..., nv2)
            Transformed data.
        """
        if normalize not in ("element", "sum", "none"):
            raise ValueError("Invalid normalization strategy.")

        mat = self.mat.T if inverse else self.mat

        if normalize == "element":
            sum_ = mat.sum(axis=0)
            scale = np.zeros_like(sum_)
            mask = sum_ != 0
            scale[mask] = 1.0 / sum_[mask]
            mat = mat @ sparse.diags(scale)
        elif normalize == "sum":
            sum_ = mat.sum(axis=1)
            scale = np.zeros_like(sum_)
            mask = sum_ != 0
            scale[mask] = 1.0 / sum_[mask]
            mat = sparse.diags(scale) @ mat

        if isinstance(x, (SurfaceMesh, pathlib.PurePath, str)):
            x = SurfaceMesh(x)
            if not x.check_sphere():
                raise ValueError("If x is a surface, it should be a sphere.")
            x.set_radius()
            rs_coords = x._coords.T @ mat

            y = SurfaceMesh.from_arrays(rs_coords.T, self.reference._triangles)
            y.set_radius()
        else:
            y = x @ mat
        return y

    def _to_hdf5(self, x5_root):
        """Write transform to HDF5 file."""

        if not _H5_AVAILABLE:
            raise ImportError("h5py")

        triangles = x5_root.create_group("Triangles")
        coords = x5_root.create_group("Coordinates")
        coords.create_dataset("0", data=self.reference._coords)
        coords.create_dataset("1", data=self.moving._coords)
        triangles.create_dataset("0", data=self.reference._triangles)
        triangles.create_dataset("1", data=self.moving._triangles)
        xform = x5_root.create_group("Transform")
        xform.attrs["Type"] = "SurfaceResampling"
        xform.attrs['InterpolationMethod'] = self.interpolation_method
        mat = xform.create_group("IndexWeights")
        mat.create_dataset("Data", data=self.mat.data)
        mat.create_dataset("Indices", data=self.mat.indices)
        mat.create_dataset("Indptr", data=self.mat.indptr)
        mat.create_dataset("Shape", data=self.mat.shape)
        reference = xform.create_group("Reference")
        reference['Coordinates'] = h5py.SoftLink('/0/Coordinates/0')
        reference['Triangles'] = h5py.SoftLink('/0/Triangles/0')
        moving = xform.create_group("Moving")
        moving['Coordinates'] = h5py.SoftLink('/0/Coordinates/1')
        moving['Triangles'] = h5py.SoftLink('/0/Triangles/1')

    def to_filename(self, filename, fmt=None):
        """Store the transform."""
        if fmt is None:
            fmt = "npz" if filename.endswith(".npz") else "X5"

        if fmt == "npz":
            raise NotImplementedError
            # sparse.save_npz(filename, self.mat)
            # return filename

        if not _H5_AVAILABLE:
            raise ImportError("h5py")

        with h5py.File(filename, "w") as out_file:
            out_file.attrs["Format"] = "X5"
            out_file.attrs["Version"] = np.uint16(1)
            root = out_file.create_group("/0")
            self._to_hdf5(root)

        return filename

    @classmethod
    def from_filename(cls, filename=None, reference_path=None, moving_path=None,
                      fmt=None, interpolation_method=None):
        """Load transform from file."""
        if filename is None:
            if reference_path is None or moving_path is None:
                raise ValueError("You must pass either a X5 file or a pair of reference and moving"
                                 " surfaces.")
            if interpolation_method is None:
                interpolation_method = 'barycentric'
            return cls(SurfaceMesh(nb.load(reference_path)),
                       SurfaceMesh(nb.load(moving_path)),
                       interpolation_method=interpolation_method)

        if fmt is None:
            try:
                fmt = "npz" if filename.endswith(".npz") else "X5"
            except AttributeError:
                fmt = "npz" if filename.as_posix().endswith(".npz") else "X5"

        if fmt == "npz":
            raise NotImplementedError
            # return cls(sparse.load_npz(filename))

        if fmt != "X5":
            raise ValueError("Only npz and X5 formats are supported.")

        if not _H5_AVAILABLE:
            raise ImportError("h5py")

        with h5py.File(filename, "r") as f:
            assert f.attrs["Format"] == "X5"
            xform = f["/0/Transform"]
            try:
                iws = xform['IndexWeights']
                mat = sparse.csr_matrix(
                    (iws["Data"][()], iws["Indices"][()], iws["Indptr"][()]),
                    shape=iws["Shape"][()],
                )
            except KeyError:
                mat = None
            reference = SurfaceMesh.from_arrays(
                xform['Reference']['Coordinates'],
                xform['Reference']['Triangles']
            )

            moving = SurfaceMesh.from_arrays(
                xform['Moving']['Coordinates'],
                xform['Moving']['Triangles']
            )
            interpolation_method = xform.attrs['InterpolationMethod']
        return cls(reference, moving, interpolation_method=interpolation_method, mat=mat)


def _points_to_triangles(points, triangles):

    """Implementation that vectorizes project of a point to a set of triangles.
    from: https://stackoverflow.com/a/32529589
    """
    with np.errstate(all='ignore'):
        # Unpack triangle points
        p0, p1, p2 = np.asarray(triangles).swapaxes(0, 1)

        # Calculate triangle edges
        e0 = p1 - p0
        e1 = p2 - p0
        a = np.einsum('...i,...i', e0, e0)
        b = np.einsum('...i,...i', e0, e1)
        c = np.einsum('...i,...i', e1, e1)

        # Calculate determinant and denominator
        det = a * c - b * b
        inv_det = 1. / det
        denom = a - 2 * b + c

        # Project to the edges
        p = p0 - points[:, np.newaxis]
        d = np.einsum('...i,...i', e0, p)
        e = np.einsum('...i,...i', e1, p)
        u = b * e - c * d
        v = b * d - a * e

        # Calculate numerators
        bd = b + d
        ce = c + e
        numer0 = (ce - bd) / denom
        numer1 = (c + e - b - d) / denom
        da = -d / a
        ec = -e / c

        # Vectorize test conditions
        m0 = u + v < det
        m1 = u < 0
        m2 = v < 0
        m3 = d < 0
        m4 = a + d > b + e

        m5 = ce > bd

        t0 = m0 & m1 & m2 & m3
        t1 = m0 & m1 & m2 & ~m3
        t2 = m0 & m1 & ~m2
        t3 = m0 & ~m1 & m2
        t4 = m0 & ~m1 & ~m2
        t5 = ~m0 & m1 & m5
        t6 = ~m0 & m1 & ~m5
        t7 = ~m0 & m2 & m4
        t8 = ~m0 & m2 & ~m4
        t9 = ~m0 & ~m1 & ~m2

        u = np.where(t0, np.clip(da, 0, 1), u)
        v = np.where(t0, 0, v)
        u = np.where(t1, 0, u)
        v = np.where(t1, 0, v)
        u = np.where(t2, 0, u)
        v = np.where(t2, np.clip(ec, 0, 1), v)
        u = np.where(t3, np.clip(da, 0, 1), u)
        v = np.where(t3, 0, v)
        u *= np.where(t4, inv_det, 1)
        v *= np.where(t4, inv_det, 1)
        u = np.where(t5, np.clip(numer0, 0, 1), u)
        v = np.where(t5, 1 - u, v)
        u = np.where(t6, 0, u)
        v = np.where(t6, 1, v)
        u = np.where(t7, np.clip(numer1, 0, 1), u)
        v = np.where(t7, 1 - u, v)
        u = np.where(t8, 1, u)
        v = np.where(t8, 0, v)
        u = np.where(t9, np.clip(numer1, 0, 1), u)
        v = np.where(t9, 1 - u, v)

        # Return closest points
        return (p0.T + u[:, np.newaxis] * e0.T + v[:, np.newaxis] * e1.T).swapaxes(2, 1)


def _barycentric_weights(vecs, coords):
    """Compute the weights for barycentric interpolation.

    Parameters
    ----------
    vecs : ndarray of shape (6, 3)
        The 6 vectors used to compute barycentric weights.
        a, e1, e2,
        np.cross(e1, e2),
        np.cross(e2, a),
        np.cross(a, e1)
    coords : ndarray of shape (3, )

    Returns
    -------
    (w, u, v, t) : tuple of float
        ``w``, ``u``, and ``v`` are the weights of the three vertices of the
        triangle, respectively. ``t`` is the scale that needs to be multiplied
        to ``coords`` to make it in the same plane as the three vertices.

    From: https://github.com/neuroboros/neuroboros/blob/\
f2a2efb914e783add2bf06e0f3715236d3d8550e/src/neuroboros/surface/_barycentric.pyx#L9-L47
    """
    det = coords[0] * vecs[3, 0] + coords[1] * vecs[3, 1] + coords[2] * vecs[3, 2]
    if det == 0:
        if vecs[3, 0] == 0 and vecs[3, 1] == 0 and vecs[3, 2] == 0:
            warnings.warn("Zero cross product of two edges: "
                          "The three vertices are in the same line.")
        else:
            print(vecs[3])
        y = coords - vecs[0]
        u, v = np.linalg.lstsq(vecs[1:3].T, y, rcond=None)[0]
        t = 1.
    else:
        uu = coords[0] * vecs[4, 0] + coords[1] * vecs[4, 1] + coords[2] * vecs[4, 2]
        vv = coords[0] * vecs[5, 0] + coords[1] * vecs[5, 1] + coords[2] * vecs[5, 2]
        u = uu / det
        v = vv / det
        tt = vecs[0, 0] * vecs[3, 0] + vecs[0, 1] * vecs[3, 1] + vecs[0, 2] * vecs[3, 2]
        t = tt / det
    w = 1. - (u + v)
    return w, u, v, t


def _find_close_tris(kdsv, tri_lut, surface):
    tris = []
    for kk in kdsv:
        tris.extend(tri_lut[kk])
    close_tri_verts = surface._triangles[np.unique(tris)]
    close_tris = surface._coords[close_tri_verts]
    return close_tris


def _find_weights(point, close_tris, d_tree):
    point = point[np.newaxis, :]
    tri_dists = cdist(point, _points_to_triangles(point, close_tris).squeeze())

    closest_tri = close_tris[(tri_dists == tri_dists.min()).squeeze()]
    # make sure a single closest triangle was found
    if closest_tri.shape[0] != 1:
        # in the event of a tie (which can happen)
        # just take the first triangle
        closest_tri = closest_tri[0]

    closest_tri = closest_tri.squeeze()
    # Make sure point is actually inside triangle
    enclosing = True
    if np.all((point > closest_tri).sum(0) != 3):

        enclosing = False
    _, ct_idxs = d_tree.query(closest_tri)
    a = closest_tri[0]
    e1 = closest_tri[1] - a
    e2 = closest_tri[2] - a
    vecs = np.vstack([a, e1, e2, np.cross(e1, e2), np.cross(e2, a), np.cross(a, e1)])
    res = {}
    res[ct_idxs[0]], res[ct_idxs[1]], res[ct_idxs[2]], _ = _barycentric_weights(
        vecs,
        point.squeeze()
    )
    return res, enclosing



# @license
# Copyright 2022 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from collections.abc import Callable
from typing import TypeVar

T = TypeVar("T")


def retry(
    func: Callable[[], T],
    max_attempts: int,
    delay: float = 0.01,
    exceptions: tuple[type[Exception], ...] = (Exception,),
) -> T:
    """Invokes `func` up to `max_attempts` times.

    Reties after a delay of `delay` if an exception in `exceptions` is raised.

    Args:
      func: Function to call.
      max_attempts: Maximum number of attempts.
      delay: Delay in seconds between attempts.
      exceptions: Exceptions upon which to retry.
    Returns:
      Result of successful call.
    Raises:
      First exception not in `exceptions`, or any exception on last attempt.
    """
    for i in range(max_attempts - 1):
        try:
            return func()
        except exceptions:
            time.sleep(delay)
    return func()



import argparse
import collections
import copy
import csv
import json
import os

import numpy as np

import neuroglancer
import neuroglancer.cli


class State:
    def __init__(self, path):
        self.path = path
        self.body_labels = {}

    def load(self):
        if os.path.exists(self.path):
            with open(self.path) as f:
                self.body_labels = dict(json.load(f))

    def save(self):
        tmp_path = self.path + ".tmp"
        with open(tmp_path, "w") as f:
            f.write(json.dumps(self.body_labels.items()))
        os.rename(tmp_path, self.path)


Body = collections.namedtuple(
    "Body", ["segment_id", "num_voxels", "bbox_start", "bbox_size"]
)


class Tool:
    def __init__(
        self, state_path, bodies, labels, segmentation_url, image_url, num_to_prefetch
    ):
        self.state = State(state_path)
        self.num_to_prefetch = num_to_prefetch
        self.viewer = neuroglancer.Viewer()
        self.bodies = bodies
        self.state.load()
        self.total_voxels = sum(x.num_voxels for x in bodies)
        self.cumulative_voxels = np.cumsum([x.num_voxels for x in bodies])

        with self.viewer.txn() as s:
            s.layers["image"] = neuroglancer.ImageLayer(source=image_url)
            s.layers["segmentation"] = neuroglancer.SegmentationLayer(
                source=segmentation_url
            )
            s.show_slices = False
            s.concurrent_downloads = 256
            s.gpu_memory_limit = 2 * 1024 * 1024 * 1024
            s.layout = "3d"

        key_bindings = [
            ["bracketleft", "prev-index"],
            ["bracketright", "next-index"],
            ["home", "first-index"],
            ["end", "last-index"],
            ["control+keys", "save"],
        ]
        label_keys = ["keyd", "keyf", "keyg", "keyh"]
        for label, label_key in zip(labels, label_keys):
            key_bindings.append([label_key, "label-%s" % label])

            def label_func(s, label=label):
                self.set_label(s, label)

            self.viewer.actions.add("label-%s" % label, label_func)
        self.viewer.actions.add("prev-index", self._prev_index)
        self.viewer.actions.add("next-index", self._next_index)
        self.viewer.actions.add("first-index", self._first_index)
        self.viewer.actions.add("last-index", self._last_index)
        self.viewer.actions.add("save", self.save)

        with self.viewer.config_state.txn() as s:
            for key, command in key_bindings:
                s.input_event_bindings.viewer[key] = command
            s.status_messages["help"] = "KEYS: " + " | ".join(
                f"{key}={command}" for key, command in key_bindings
            )

        self.index = -1
        self.set_index(self._find_one_after_last_labeled_index())

    def _find_one_after_last_labeled_index(self):
        body_index = 0
        while self.bodies[body_index].segment_id in self.state.body_labels:
            body_index += 1
        return body_index

    def set_index(self, index):
        if index == self.index:
            return
        body = self.bodies[index]
        self.index = index

        def modify_state_for_body(s, body):
            s.layers["segmentation"].segments = frozenset([body.segment_id])
            s.voxel_coordinates = body.bbox_start + body.bbox_size // 2

        with self.viewer.txn() as s:
            modify_state_for_body(s, body)

        prefetch_states = []
        for i in range(self.num_to_prefetch):
            prefetch_index = self.index + i + 1
            if prefetch_index >= len(self.bodies):
                break
            prefetch_state = copy.deepcopy(self.viewer.state)
            prefetch_state.layout = "3d"
            modify_state_for_body(prefetch_state, self.bodies[prefetch_index])
            prefetch_states.append(prefetch_state)

        with self.viewer.config_state.txn() as s:
            s.prefetch = [
                neuroglancer.PrefetchState(state=prefetch_state, priority=-i)
                for i, prefetch_state in enumerate(prefetch_states)
            ]

        label = self.state.body_labels.get(body.segment_id, "")
        with self.viewer.config_state.txn() as s:
            s.status_messages["status"] = (
                "[Segment %d/%d  : %d/%d voxels labeled = %.3f fraction] label=%s"
                % (
                    index,
                    len(self.bodies),
                    self.cumulative_voxels[index],
                    self.total_voxels,
                    self.cumulative_voxels[index] / self.total_voxels,
                    label,
                )
            )

    def save(self, s):
        self.state.save()

    def set_label(self, s, label):
        self.state.body_labels[self.bodies[self.index].segment_id] = label
        self.set_index(self.index + 1)

    def _first_index(self, s):
        self.set_index(0)

    def _last_index(self, s):
        self.set_index(max(0, self._find_one_after_last_labeled_index() - 1))

    def _next_index(self, s):
        self.set_index(self.index + 1)

    def _prev_index(self, s):
        self.set_index(max(0, self.index - 1))


if __name__ == "__main__":
    ap = argparse.ArgumentParser()
    neuroglancer.cli.add_server_arguments(ap)
    ap.add_argument(
        "--image-url", required=True, help="Neuroglancer data source URL for image"
    )
    ap.add_argument(
        "--segmentation-url",
        required=True,
        help="Neuroglancer data source URL for segmentation",
    )
    ap.add_argument("--state", required=True, help="Path to proofreading state file")
    ap.add_argument(
        "--bodies", required=True, help="Path to list of bodies to proofread"
    )
    ap.add_argument("--labels", nargs="+", help="Labels to use")
    ap.add_argument(
        "--prefetch", type=int, default=10, help="Number of bodies to prefetch"
    )

    args = ap.parse_args()
    neuroglancer.cli.handle_server_arguments(args)

    bodies = []

    with open(args.bodies) as f:
        csv_reader = csv.DictReader(f)
        for row in csv_reader:
            bodies.append(
                Body(
                    segment_id=int(row["id"]),
                    num_voxels=int(row["num_voxels"]),
                    bbox_start=np.array(
                        [
                            int(row["bbox.start.x"]),
                            int(row["bbox.start.y"]),
                            int(row["bbox.start.z"]),
                        ],
                        dtype=np.int64,
                    ),
                    bbox_size=np.array(
                        [
                            int(row["bbox.size.x"]),
                            int(row["bbox.size.y"]),
                            int(row["bbox.size.z"]),
                        ],
                        dtype=np.int64,
                    ),
                )
            )

    tool = Tool(
        state_path=args.state,
        image_url=args.image_url,
        segmentation_url=args.segmentation_url,
        labels=args.labels,
        bodies=bodies,
        num_to_prefetch=args.prefetch,
    )
    print(tool.viewer)



"""Various utilities."""
# stdlib
import json
import logging
import socket
import sys
from urllib.parse import quote

# optionals
try:
    import neuroglancer as ng
except ImportError:
    import ngtools._nglite as ng

LOG = logging.getLogger(__name__)


NG_URLS = {
    "ng":
        "https://neuroglancer-demo.appspot.com/",
    "linc":
        "https://neuroglancer.lincbrain.org/cloudfront/frontend/index.html",
}


def neuroglancer_state_to_neuroglancer_url(
    state: ng.ViewerState,
    base_url: str = NG_URLS["ng"],
) -> str:
    """
    Take the current state of the Neuroglancer viewer and convert it
    into a url-encoded value for the browser.
    """
    ordered_dict = state.to_json()

    # Convert the dictionary to a JSON string
    json_str = json.dumps(ordered_dict)

    # URL-encode the JSON string
    encoded_json = quote(json_str)

    # Construct and return the full URL
    return f"{base_url}#!{encoded_json}"


_NG_TYPE_WRAPPERS = {}


def Wraps(kls: type) -> type:
    """
    Return a wrapper type that inherits from `kls` but never calls
    `super().__init__()`.

    Instead, it stores an instance of the parent class and overloads
    attribute access so that attributes of the parent instance -- rather
    than `self` -- are accessed and set.

    This is necessary when wrapping neuroglancer types into our own types.
    """
    if kls not in _NG_TYPE_WRAPPERS:

        class _Wraps(kls):

            __name__ = "Wrapped" + kls.__name__
            __doc__ = kls.__doc__

            def __init__(self, *args, **kwargs) -> None:
                # if there is a single argument and it has the right
                # type, just wrap it (no deep copy)
                if len(args) == 1 and isinstance(args[0], kls) and not kwargs:
                    wrapped = args[0]
                    if type(wrapped) is _NG_TYPE_WRAPPERS[kls]:
                        wrapped = wrapped._wrapped
                # otherwise, call the wraped class constructor (deep copy)
                else:
                    wrapped = kls(*args, **kwargs)
                object.__setattr__(self, "_wrapped", wrapped)

            def __fix_none_value__(self, name: str, value: object) -> object:
                # magic method for attributes that have a __default__
                if name.startswith("__"):
                    return value
                if value is None:
                    value = getattr(self, f"__default_{name}__", value)
                return value

            def __getattribute__(self, name: str) -> object:
                # FIXME: I haven't managed to use __getattr__ only,
                # as it fails when the wrapped object implements its
                # own fancy __getattr__ (e.g. `ManagedLayer`).
                # It *might* be possible to solve this, but in the
                # meantime I had to overload __getattribute__. This
                # is annoying, as overloading __getattribute__ has
                # a non-zero performance cost.

                # magic objects -- never defer to _wrapped
                if name.startswith("__") or name == "_wrapped":
                    return object.__getattribute__(self, name)

                # check if a getter is available
                try:
                    value = object.__getattribute__(self, f"__get_{name}__")()
                    return self.__fix_none_value__(name, value)
                except AttributeError:
                    ...

                # check if attribute exists in _wrapped
                if name != "_wrapped" and "_wrapped" in self.__dict__:
                    try:
                        value = self._wrapped.__getattribute__(name)
                        return self.__fix_none_value__(name, value)
                    except AttributeError:
                        ...

                # fallback to default implementation
                value = object.__getattribute__(self, name)
                return self.__fix_none_value__(name, value)

            def __getattr__(self, name: str) -> object | type:
                if name in self.__dict__:
                    # FIXME: I don't think we ever enter this branch
                    # -> should be dealt with in __getattribute__
                    value = self.__dict__[name]
                    return self.__fix_none_value__(name, value)
                if "_wrapped" in self.__dict__:
                    # FIXME: not entirely sure I need this test -- I
                    # should be able to assume _wrapped is available.
                    # And if not, we're happy with it raising an
                    # AttributeError anyway.
                    value = getattr(self._wrapped, name)
                    return self.__fix_none_value__(name, value)
                raise AttributeError(name)

            def __setattr__(self, name: str, value: object | type) -> None:
                if value is None:
                    value = self.__fix_none_value__(name, value)
                # check if a setter is available
                if hasattr(self, f"__set_{name}__"):
                    value = getattr(self, f"__set_{name}__")(value)
                # if attribute exists in the main object, set it there
                if name in self.__dict__:
                    # NOTE: I don't remember why I had to use the
                    # object method here.
                    return object.__setattr__(self, name, value)
                # if attribute exists in the warpped object, set it there
                if "_wrapped" in self.__dict__:
                    if hasattr(self._wrapped, name):
                        return setattr(self._wrapped, name, value)
                # fallback to default implementation.
                return super().__setattr__(name, value)

            def __delattr__(self, name: str) -> None:
                # if attribute exists in the main object, delete it there
                if name in self.__dict__:
                    return super().__delattr__(name)
                # if attribute exists in the wrapped object, delete it there
                if hasattr(self, "_wrapped") and hasattr(self._wrapped, name):
                    return delattr(self._wrapped, name)
                # fallback to default implementation.
                return super().__delattr__(name)

        _Wraps.__wrapper_class__ = _Wraps
        _Wraps.__wrapped_class__ = kls
        _NG_TYPE_WRAPPERS[kls] = _Wraps

    return _NG_TYPE_WRAPPERS[kls]


def find_available_port(port: int = 0, ip: str = "") -> tuple[int, str]:
    """Return an available port and the local IP."""
    try:
        s = socket.socket()
        s.bind((ip, port))
        ip = s.getsockname()[0]
        port = s.getsockname()[1]
        s.close()
    except OSError:
        port0 = port
        s = socket.socket()
        s.bind((ip, 0))
        ip = s.getsockname()[0]
        port = s.getsockname()[1]
        s.close()
        print(f'Port {port0} already in use. Use port {port} instead.',
              file=sys.stderr)
    return port, ip
